#!/usr/bin/env python

""" Test Mapping utility functions
"""

import doctest
import os
import re
from collections import defaultdict
from pathlib import Path
from typing import List
from unittest import TestCase, main

import yaml
from pep440.core import is_canonical

import g2p
import g2p.exceptions
from g2p import get_arpabet_langs
from g2p._version import VERSION, version_tuple
from g2p.log import LOGGER
from g2p.mappings import Mapping, utils
from g2p.mappings.utils import RULE_ORDERING_ENUM, Rule
from g2p.tests.public import PUBLIC_DIR


class UtilsTest(TestCase):
    def test_run_doctest(self):
        """Run doctests in g2p.mappings.utils"""
        results = doctest.testmod(utils)
        self.assertFalse(results.failed, results)

    def test_abb_expand(self):
        test_dict = defaultdict(list)
        bad_dict = defaultdict(list)
        test_dict["VOWELS"].extend(["HIGH_VOWELS", "e", "o"])
        test_dict["HIGH_VOWELS"].extend(["i", "u"])
        bad_dict["VOWELS"].extend(["HIGH_VOWELS", "e", "o"])
        bad_dict["HIGH_VOWELS"].extend(
            ["HIGH_VOWELS", "u"]
        )  # shouldn't allow self-referential abbreviations
        expanded_plain = utils.expand_abbreviations("test", test_dict)
        expanded_bad_plain = utils.expand_abbreviations("test", bad_dict)
        with self.assertRaises(g2p.exceptions.RecursionError):
            utils.expand_abbreviations("HIGH_VOWELS", bad_dict)
        expanded_non_recursive = utils.expand_abbreviations("HIGH_VOWELS", test_dict)
        expanded_recursive = utils.expand_abbreviations("VOWELS", test_dict)
        self.assertEqual("test", expanded_plain)
        self.assertEqual("test", expanded_bad_plain)
        self.assertEqual("i|u", expanded_non_recursive)
        self.assertEqual("i|u|e|o", expanded_recursive)

    def test_abb_flatten_and_expand_format(self):
        test_rows = [["VOWEL", "a", "e", "i", "o", "u"], ["OTHER", "t", "e", "s", "t"]]
        default_dict = defaultdict(list)
        default_dict["VOWEL"].extend(["a", "e", "i", "o", "u"])
        default_dict["OTHER"].extend(["t", "e", "s", "t"])
        empty_rows: List[List[str]] = []
        while len(empty_rows) < 10:
            empty_rows.append(["", "", "", "", "", ""])
        self.assertEqual(utils.flatten_abbreviations_format(test_rows), default_dict)
        self.assertEqual(utils.expand_abbreviations_format(default_dict), test_rows)
        self.assertEqual(utils.expand_abbreviations_format({}), empty_rows)

    def test_unicode_escape(self):
        """Should turn \u0331 declared in CSVs
        into actual Unicode string for that codepoint
        """
        self.assertEqual("\u0000", utils.unicode_escape("\\u0000"))
        self.assertEqual("\u0331", utils.unicode_escape("\\u0331"))
        self.assertEqual("\u26F0", utils.unicode_escape("\\u26F0"))

    def test_fixed_width(self):
        test_dict = defaultdict(list)
        test_dict["VOWELS"].extend(["e", "o", "ee"])
        lookbehind_pattern = re.compile(r"\(\?\<\=[^)]*\)")
        patterns = [
            (utils.create_fixed_width_lookbehind("a|b"), 1),
            (utils.create_fixed_width_lookbehind("a|b|cc"), 2),
            (utils.create_fixed_width_lookbehind("a|'|b|cc|ddd|$"), 4),
            (utils.create_fixed_width_lookbehind("a|^|$"), 2),
            (utils.create_fixed_width_lookbehind("[abcd]"), 1),
            (utils.create_fixed_width_lookbehind(r"[x'kgh\.𝚐̲𝚔̲𝚡̲̲]"), 1),
            (
                utils.create_fixed_width_lookbehind(
                    utils.expand_abbreviations("VOWELS", test_dict)
                ),
                2,
            ),
            (
                utils.create_fixed_width_lookbehind(
                    utils.expand_abbreviations("(VOWELS|eee)", test_dict)
                ),
                3,
            ),
        ]
        for pattern in patterns:
            self.assertEqual(
                len(re.split(lookbehind_pattern, pattern[0])) - 1, pattern[1]
            )

    def test_load_mapping(self):
        with self.assertLogs(LOGGER, "WARNING"):
            Mapping.load_mapping_from_path(
                os.path.join(PUBLIC_DIR, "mappings", "malformed_config-g2p.yaml")
            )
        minimal = Mapping.load_mapping_from_path(
            os.path.join(PUBLIC_DIR, "mappings", "minimal_config-g2p.yaml")
        )
        csv = Mapping.load_mapping_from_path(
            os.path.join(PUBLIC_DIR, "mappings", "minimal_configs.yaml"), 0
        )
        tsv = Mapping.load_mapping_from_path(
            os.path.join(PUBLIC_DIR, "mappings", "minimal_configs.yaml"), 1
        )
        psv = Mapping.load_mapping_from_path(
            os.path.join(PUBLIC_DIR, "mappings", "minimal_configs.yaml"), 2
        )
        json = Mapping.load_mapping_from_path(
            os.path.join(PUBLIC_DIR, "mappings", "minimal_configs.yaml"), 3
        )
        xlsx = Mapping.load_mapping_from_path(
            os.path.join(PUBLIC_DIR, "mappings", "minimal_configs.yaml"), 4
        )
        self.assertEqual(minimal.rules, csv.rules)
        self.assertEqual(minimal.rules, tsv.rules)
        self.assertEqual(minimal.rules, psv.rules)
        self.assertEqual(minimal.rules, json.rules)
        self.assertEqual(minimal.rules, xlsx.rules)

    def test_escape_special(self):
        self.assertEqual(
            utils.escape_special_characters(
                Rule(rule_input="?", rule_output="")
            ).rule_input,
            "\\?",
        )

    def test_load_abbs(self):
        with self.assertRaises(g2p.exceptions.IncorrectFileType):
            utils.load_abbreviations_from_file(
                os.path.join(PUBLIC_DIR, "mappings", "abbreviations.json")
            )
        for abb in ["abbreviations.csv", "abbreviations.tsv", "abbreviations.psv"]:
            abbs = utils.load_abbreviations_from_file(
                os.path.join(PUBLIC_DIR, "mappings", abb)
            )
            self.assertTrue("VOWEL" in abbs)
            self.assertEqual(abbs["VOWEL"], ["a", "e", "i", "o", "u"])

    def test_generated_mapping(self):
        try:
            # config = utils.generate_config('test', 'test-out', 'Test', 'TestOut')
            mapping = Mapping(
                in_lang="test",
                out_lang="test-out",
                rule_ordering=RULE_ORDERING_ENUM.apply_longest_first,
                rules=[Rule(rule_input="a", rule_output="b")],
            )
            with self.assertLogs(LOGGER, level="WARNING"):
                mapping.config_to_file(
                    os.path.join(PUBLIC_DIR, "mappings", "test_config-g2p.yaml")
                )
            with self.assertLogs(LOGGER, level="WARNING"):
                mapping.config_to_file(
                    os.path.join(PUBLIC_DIR, "mappings", "generated_add.yaml")
                )
            mapping.mapping_to_file(os.path.join(PUBLIC_DIR, "mappings"))
            test_config = Mapping.load_mapping_from_path(
                os.path.join(PUBLIC_DIR, "mappings", "test_config-g2p.yaml")
            )

            test_config_added = Mapping.load_mapping_from_path(
                os.path.join(PUBLIC_DIR, "mappings", "generated_add.yaml")
            )
            self.assertEqual(
                test_config.rules[0].export_to_dict(),
                Rule(
                    **{"in": "a", "out": "b", "context_before": "", "context_after": ""}
                ).export_to_dict(),
            )
            self.assertEqual(test_config.in_lang, "test")
            self.assertEqual(test_config.out_lang, "test-out")
            self.assertEqual(test_config.language_name, "test")
            self.assertEqual(test_config.display_name, "test custom to test-out custom")
            self.assertEqual(
                test_config_added.rules[0].export_to_dict(),
                {"in": "a", "out": "b"},
            )
            self.assertEqual(test_config_added.in_lang, "test")
            self.assertEqual(test_config_added.out_lang, "test-out")
            self.assertEqual(test_config_added.language_name, "test")
            self.assertEqual(
                test_config_added.display_name, "test custom to test-out custom"
            )
        finally:
            gen_mapping = os.path.join(PUBLIC_DIR, "mappings", "test_to_test-out.json")
            gen_config = os.path.join(PUBLIC_DIR, "mappings", "test_config-g2p.yaml")
            if os.path.exists(gen_config):
                os.remove(gen_config)
            if os.path.exists(gen_mapping):
                os.remove(gen_mapping)
            fresh_config = {"language_name": "generated", "mappings": []}
            with open(
                os.path.join(PUBLIC_DIR, "mappings", "generated_add.yaml"),
                "w",
                encoding="utf8",
            ) as f:
                yaml.dump(
                    fresh_config, f, Dumper=utils.IndentDumper, default_flow_style=False
                )

    def test_bad_normalization(self):
        with self.assertRaises(g2p.exceptions.InvalidNormalization):
            utils.normalize_with_indices("test", "bad")

    def test_normalize_to_NFD_with_indices(self):
        # Useful site to get combining character code points:
        # http://www.alanwood.net/unicode/combining_diacritical_marks.html
        e_acute_nfd = "e\u0301"
        self.assertEqual(
            utils.normalize_with_indices("é", "NFD"),
            (e_acute_nfd, [(0, 0), (0, 1)]),
        )
        o_graveabove_nfd = "o\u0300"
        self.assertEqual(
            utils.normalize_with_indices("ò", "NFD"),
            (o_graveabove_nfd, [(0, 0), (0, 1)]),
        )
        # TODO: this test case really should have indices (0,0),(0,2), (1,1)
        o_graveabove_acutebelow_mixed = "ò\u0317"  # 'ò̗'
        o_graveabove_acutebelow_nfd = "o\u0317\u0300"  # 'ò̗'
        self.assertEqual(
            utils.normalize_with_indices(o_graveabove_acutebelow_mixed, "NFD"),
            (o_graveabove_acutebelow_nfd, [(0, 0), (0, 2), (1, 1)]),
        )
        o_graveabove_acutebelow_disordered = "o\u0300\u0317"
        self.assertEqual(
            utils.normalize_with_indices(o_graveabove_acutebelow_disordered, "NFD"),
            (o_graveabove_acutebelow_nfd, [(0, 0), (1, 2), (2, 1)]),
        )
        # From https://en.wikipedia.org/wiki/Precomposed_character:
        # "\u1e53" (ṓ) == "\u014d\u0301" (ṓ) == "\u006f\u0304\u0301" (ṓ)
        self.assertEqual(
            utils.normalize_with_indices("\u1e53", "NFD"),
            ("\u006f\u0304\u0301", [(0, 0), (0, 1), (0, 2)]),
        )
        self.assertEqual(
            utils.normalize_with_indices("\u014d\u0301", "NFD"),
            ("\u006f\u0304\u0301", [(0, 0), (0, 1), (1, 2)]),
        )
        self.assertEqual(
            utils.normalize_with_indices("'שָׂ'", "NFD"),
            ("'שָׂ'", [(0, 0), (1, 1), (2, 3), (3, 2), (4, 4)]),
        )

    def test_compose_indices(self):
        self.assertEqual(
            utils.compose_indices([(0, 1), (1, 4)], [(0, 0), (1, 2), (1, 3), (4, 2)]),
            [(0, 2), (0, 3), (1, 2)],
        )
        self.assertEqual(
            utils.compose_indices([(0, 0), (0, 1), (1, 2)], [(0, 3), (1, 3), (2, 3)]),
            [(0, 3), (1, 3)],
        )
        self.assertEqual(
            utils.compose_indices([(0, 1), (1, 2)], [(1, 4), (3, 1)]),
            [(0, 4)],
        )

    def test_normalize_to_NFC_with_indices(self):
        self.assertEqual(
            utils.normalize_with_indices("e\u0301", "NFC"),
            ("é", [(0, 0), (1, 0)]),
        )
        self.assertEqual(
            utils.normalize_with_indices("ò\u0317", "NFC"),
            ("ò̗", [(0, 0), (1, 1)]),
        )
        self.assertEqual(
            utils.normalize_with_indices("\u014d\u0301", "NFC"),
            ("\u1e53", [(0, 0), (1, 0)]),
        )
        self.assertEqual(
            utils.normalize_with_indices("o\u0304\u0301", "NFC"),
            ("\u1e53", [(0, 0), (1, 0), (2, 0)]),
        )
        self.assertEqual(
            utils.normalize_with_indices("\u014d\u0301", "none"),
            ("\u014d\u0301", [(0, 0), (1, 1)]),
        )
        self.assertEqual(
            utils.normalize_with_indices("o\u0300\u0317", "NFC"),
            ("\u00F2\u0317", [(0, 0), (1, 0), (2, 1)]),
        )

    def test_normalize_to_NFK_with_indices(self):
        e_acute_nfd = "e\u0301"
        self.assertEqual(
            utils.normalize_with_indices(e_acute_nfd, "NFKC"),
            ("é", [(0, 0), (1, 0)]),
        )
        self.assertEqual(
            utils.normalize_with_indices("é", "NFKD"),
            (e_acute_nfd, [(0, 0), (0, 1)]),
        )

    def test_get_arpabet_langs(self):
        LANGS, LANG_NAMES = get_arpabet_langs()
        self.assertEqual(LANGS, sorted(LANGS))
        self.assertEqual(list(LANG_NAMES.keys()), sorted(LANG_NAMES.keys()))
        self.assertEqual(LANGS, list(LANG_NAMES.keys()))
        self.assertTrue("kwk-umista" in LANG_NAMES)
        self.assertTrue("str" in LANG_NAMES)
        self.assertGreater(len(LANGS), 40)
        LANGS2, LANG_NAMES2 = get_arpabet_langs()
        self.assertIs(LANGS2, LANGS)
        self.assertIs(LANG_NAMES2, LANG_NAMES)

    def test_version_is_pep440_compliant(self):
        """We test for almost PEP 440 compliance: hatch adds +local_sha1, which is not compliant."""
        main_version, _, _ = VERSION.partition("+")
        self.assertTrue(is_canonical(main_version))

    def test_scm_pretend_version_is_up_to_date(self):
        """.SETUPTOOLS_SCM_PRETEND_VERSION is set to the version in pyproject.toml"""
        filename = Path(g2p.__file__).parent.parent / ".SETUPTOOLS_SCM_PRETEND_VERSION"
        try:
            with open(filename) as f:
                pretend_version = f.read().strip()
            (major, minor, *_rest) = version_tuple
            major_minor = f"{major}.{minor}"
            self.assertEqual(
                major_minor,
                pretend_version,
                "Mismatch between .SETUPTOOLS_SCM_PRETEND_VERSION and the version setuptools_scm determined dynamically. Try: 1) fetch recent tags from GitHub, 2) rerun \"pip install -e .\", 3) if you're working on the next major or minor release, update .SETUPTOOLS_SCM_PRETEND_VERSION to match the dynamic version's major.minor.",
            )
        except FileNotFoundError:
            # This is fine, it's only used in development
            pass

    def test_token_class(self):
        from g2p.shared_types import Token

        t1 = Token("test", True)
        t2 = Token(":", False)

        # Current and deprecated usages
        for t in t1, t2:
            with self.assertWarns(DeprecationWarning):
                self.assertEqual(t.text, t["text"])
            with self.assertWarns(DeprecationWarning):
                self.assertEqual(t.is_word, t["is_word"])
        # new way to set
        t1.text = "test2"
        t1.is_word = False
        self.assertEqual(t1.text, "test2")
        self.assertEqual(t1.is_word, False)
        # deprecated way to set
        with self.assertWarns(DeprecationWarning):
            t1["text"] = "test3"
        with self.assertWarns(DeprecationWarning):
            t1["is_word"] = True
        self.assertEqual(t1.text, "test3")
        self.assertEqual(t1.is_word, True)

        with self.assertRaises(KeyError):
            with self.assertWarns(DeprecationWarning):
                t1["bad_key"] = "test"
        with self.assertRaises(KeyError):
            with self.assertWarns(DeprecationWarning):
                _ = t2["bad_key"]


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import json

with open("eng_ipa_to_arpabet.json") as f:
    mappings = json.load(f)
    print("[")
    seen = set()
    for i, m in enumerate(mappings):
        if m["out"] in seen:
            continue
        seen.add(m["out"])
        if " " in m["out"]:
            continue
        print(
            '    { "in": "%s", "out": "%s" }%s'
            % (m["out"], m["in"], "," if i != len(mappings) - 1 else "")
        )
    print("]")



"""
This file is for abstract type definitions that can be initialized without any
(expensive) dependencies.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List

from typing_extensions import deprecated


@dataclass
class Token:
    """A token from the g2p tokenizer."""

    text: str
    is_word: bool

    @deprecated(
        "Accessing g2p Token objects as dicts is deprecated since g2p 2.2.0. "
        "Please use the 'text' and 'is_word' attributes instead.",
    )
    def __getitem__(self, key):
        """For backward compatibility only, allow access as if it were a dict."""
        if key == "text":
            return self.text
        if key == "is_word":
            return self.is_word
        raise KeyError(key)

    @deprecated(
        "Accessing g2p Token objects as dicts is deprecated since g2p 2.2.0. "
        "Please use the 'text' and 'is_word' attributes instead.",
    )
    def __setitem__(self, key, value):
        """For backward compatibility only, allow setting values as if it were a dict."""
        if key == "text":
            self.text = value
        elif key == "is_word":
            self.is_word = value
        else:
            raise KeyError(key)


class BaseTransducer(ABC):
    """Base class to typecheck transducers without having to import them."""

    @abstractmethod
    def __call__(self, to_convert: str):
        """Transduce to_convert."""

    @property
    @abstractmethod
    def transducers(self):
        """A list of BaseTransducer objects for each tier in the transducer."""

    @property
    @abstractmethod
    def in_lang(self) -> str:
        """The input language code of the transducer."""

    @property
    @abstractmethod
    def out_lang(self) -> str:
        """The output language code of the transducer."""

    @abstractmethod
    def check(
        self, tg: "BaseTransductionGraph", shallow=False, display_warnings=False
    ) -> bool:
        """Check if the output is valid in the target symbol space"""


class BaseTransductionGraph(ABC):
    """Base class to typecheck transduction graphs without having to import them."""

    @property
    @abstractmethod
    def tiers(self):
        """A list of BaseTransductionGraph objects for each tier in the graph."""


class BaseTokenizer(ABC):
    """Base class to typecheck tokenizers without having to import them."""

    @abstractmethod
    def tokenize_text(self, text: str) -> List[Token]:
        """Tokenize text."""



"""

Utilities used by other classes

"""

import csv
import functools
import json
import os
import unicodedata as ud
from bisect import bisect_left
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import regex as re
import yaml
from pydantic import (
    BaseModel,
    ConfigDict,
    DirectoryPath,
    Field,
    ValidationInfo,
    field_serializer,
    field_validator,
    model_validator,
)
from typing_extensions import Literal

from g2p import exceptions
from g2p.exceptions import NeuralDependencyError
from g2p.log import LOGGER
from g2p.mappings import langs
from g2p.shared_types import Token

GEN_DIR = os.path.join(os.path.dirname(langs.__file__), "generated")
GEN_CONFIG = os.path.join(GEN_DIR, "config-g2p.yaml")


def has_neural_support() -> bool:
    """Check if neural optional dependencies are installed."""
    try:  # type: ignore
        import deep_phonemizer  # type: ignore[import-not-found]  # noqa: F401
        import huggingface_hub  # type: ignore[import-not-found]  # noqa: F401
        import torch  # type: ignore[import-not-found]  # noqa: F401

        return True
    except ImportError:
        return False


def requires_neural(func):
    """Decorator to ensure neural dependencies are available."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not has_neural_support():
            raise NeuralDependencyError(func.__name__)
        return func(*args, **kwargs)

    return wrapper


@requires_neural
def deep_phonemizer_handler(
    model_path: Path,
    config_kwargs: dict,
    to_convert: str,
) -> str:
    from deep_phonemizer.phonemizer import Phonemizer  # type: ignore[import-not-found]

    from g2p.mappings import Mapping
    from g2p.transducer import Transducer

    phonemizer = Phonemizer.from_checkpoint(model_path)
    to_convert_list = to_convert.split()
    result = phonemizer.phonemise_list(
        to_convert_list, lang=config_kwargs.get("lang", "default")
    )
    phonemized_string = " ".join(result.phonemes)
    normalizer = Transducer(Mapping.find_mapping_by_id("ipa_normalization"))
    return normalizer(phonemized_string).output_string


@requires_neural
def download_huggingface_model(
    model_id: str,
    local_dir: Optional[Union[str, Path]] = None,
    filename: Optional[str] = None,
    cache_dir: Optional[Union[str, Path]] = None,
    revision: str = "main",
    force_download: bool = False,
) -> Path:
    """
    Download a HuggingFace model or specific file from a model repository.

    Args:
        model_id: HuggingFace model identifier (e.g., "LOBOT-PENAC/g2p")
        local_dir: Local directory to save the model. If None, uses HF cache.
        filename: Specific file within the repo to download. If None, downloads entire repo.
        cache_dir: Custom cache directory. If None, uses default HF cache.
        revision: Git revision (branch, tag, or commit) to download from.
        force_download: Whether to force re-download even if cached.

    Returns:
        Path to the downloaded model/file.

    Raises:
        ImportError: If huggingface_hub is not installed.
        Exception: If download fails.
    """
    from huggingface_hub import (
        hf_hub_download,
        snapshot_download,
        try_to_load_from_cache,
    )

    # We add no-cover pragmas below because we don't want the coverage report to
    # ping-pong between cache hits and cache misses, which would report spurious
    # changes to coverage.
    # if the model is already cached, just return that
    if filename:
        cached = try_to_load_from_cache(repo_id=model_id, filename=filename)
        if not force_download and cached:  # pragma: no cover
            return cached

    try:  # pragma: no cover
        if filename is not None:
            # Download specific file
            downloaded_path = hf_hub_download(
                repo_id=model_id,
                filename=filename,
                local_dir=local_dir,
                cache_dir=cache_dir,
                revision=revision,
                force_download=force_download,
            )
            return Path(downloaded_path)
        else:
            # Download entire repository
            downloaded_path = snapshot_download(
                repo_id=model_id,
                local_dir=local_dir,
                cache_dir=cache_dir,
                revision=revision,
                force_download=force_download,
            )
            return Path(downloaded_path)

    except Exception as e:
        raise Exception(f"Failed to download {model_id}: {str(e)}")


class Rule(BaseModel):
    # We can't just use "in" because it's disallowed by Python
    rule_input: str = Field(alias="in")
    """The character(s) to convert"""

    rule_output: str = Field(alias="out")
    """What to convert the 'in' characters to"""

    context_before: str = ""
    """The context before 'in' required for the rule to apply"""

    context_after: str = ""
    """The context after 'in' required for the rule to apply"""

    prevent_feeding: bool = False
    """Whether to prevent the rule from feeding other rules"""

    match_pattern: Optional[Pattern] = Field(
        None,
        exclude=True,
        # Don't include this in the docs because it's generated, and would require a schema update
        # description="""An automatically generated match_pattern based on the rule_input, context_before and context_after""",
    )

    intermediate_form: Optional[str] = Field(
        None,
        exclude=True,
        # Don't include this in the docs because it's generated, and would require a schema update
        # description="""An intermediate form, automatically generated only when prevent_feeding is True""",
    )
    comment: Optional[str] = None
    """An optional comment about the rule."""

    model_config = ConfigDict(extra="ignore", populate_by_name=True)

    def export_to_dict(
        self, exclude=None, exclude_none=True, exclude_defaults=True, by_alias=True
    ):
        """All the options for exporting are tedious to keep track of so this is a helper function"""
        return self.model_dump(
            exclude=exclude,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
            by_alias=by_alias,
        )


def expand_abbreviations(data: str, abbs: Dict[str, List[str]], recursion_depth=0):
    """Given a string, expand any abbreviations in it recursively

    Args:
        data (str): a string that may or may not contain recursive 'abbreviations'
    """
    if recursion_depth > 10:
        raise exceptions.RecursionError(
            "Too many levels of recursion in your abbreviation expansion. "
            "Check your abbreviations for circular references."
        )
    for abb, stands_for in sorted(abbs.items(), key=lambda x: len(x[0]), reverse=True):
        abb_match = re.compile(abb)
        abb_repl = "|".join(stands_for)
        if re.search(abb_match, data):
            data = re.sub(abb_match, abb_repl, data)
            recursion_depth += 1
            data = expand_abbreviations(data, abbs, recursion_depth=recursion_depth)
            break
    return data


def flatten_abbreviations_format(data):
    """Turn a CSV-sourced list of lists into a flattened DefaultDict"""
    default_dict = defaultdict(list)
    for line in data:
        if line[0]:
            default_dict[line[0]].extend(
                [definition for definition in line[1:] if definition]
            )
    return default_dict


def expand_abbreviations_format(data):
    """Expand a flattened DefaultDict into a CSV-formatted list of lists"""
    lines = []
    if data:
        for key in data.keys():
            line = [key]
            for col in data[key]:
                line.append(col)
            lines.append(line)
    if not lines:
        while len(lines) < 10:
            lines.append(["", "", "", "", "", ""])
    return lines


def normalize(inp: str, norm_form: Union[str, None]):
    """Normalize to NFC(omposed) or NFD(ecomposed).

    Also, find any Unicode Escapes & decode 'em!
    """
    if norm_form is None or norm_form == "none":
        return unicode_escape(inp)
    if norm_form not in ["NFC", "NFD", "NFKC", "NFKD"]:
        raise exceptions.InvalidNormalization(norm_form)
    # Sadly mypy doesn't do narrowing to literals properly
    norm_form = cast(Literal["NFC", "NFD", "NFKC", "NFKD"], norm_form)
    normalized = ud.normalize(norm_form, unicode_escape(inp))
    if normalized != inp:
        LOGGER.debug(
            "The string %s was normalized to %s using the %s standard and by decoding any Unicode escapes. "
            "Note that this is not necessarily the final stage of normalization.",
            inp,
            normalized,
            norm_form,
        )
    return normalized


# compose_indices is generic because we would like to propagate the
# type of its second input, in the case where we *know* there will not
# be None (NFC and NFD conversions)
IntOrOptionalInt = TypeVar("IntOrOptionalInt", bound=Union[int, None])


def compose_indices(
    indices1: List[Tuple[int, int]], indices2: List[Tuple[int, IntOrOptionalInt]]
) -> List[Tuple[int, IntOrOptionalInt]]:
    """Compose indices1 + indices2 into direct arcs from the inputs of indices1
    to the outputs of indices 2.

    >>> compose_indices([(0,1), (1,4)], [(0,0), (1,2), (1,3), (4,2)])
    [(0, 2), (0, 3), (1, 2)]
    """
    # for O(1) lookup of arcs leaving indices2
    indices2_as_dict = defaultdict(dict)  # type: ignore
    for a, b in indices2:
        indices2_as_dict[a][b] = True  # we're using dict as an ordered set...

    result = ((a, c) for a, b in indices1 for c in indices2_as_dict[b].keys())
    return list(dict.fromkeys(result).keys())  # return a deduplicated list


def normalize_to_NFD_with_indices(
    inp: str, norm_form: str
) -> Tuple[str, List[Tuple[int, int]]]:
    """Normalize to NFD and return the indices mapping input to output characters"""
    assert norm_form in ("NFD", "NFKD")
    # Sadly mypy doesn't do narrowing to literals properly
    norm_form = cast(Literal["NFD", "NFKD"], norm_form)
    # decomposed is the valid NF(K)D decomposition, taking into account possible
    # diacritic reordering
    decomposed = ud.normalize(norm_form, inp)
    # decom
    decomp_to_use = list(enumerate(decomposed))
    indices = []
    # To figure out indices from input characters, decompose them one at a time and
    # look for each character in the jointly decomposed output, allowing for
    # the potential reordering of
    #  "Mn==Nonspacing_Mark==a nonspacing combining mark (zero advance width)"
    # characters as idenfied by unicodedata.category().
    # This handles this situation (I==input, L==letter, D==diacritic):
    #  - input is I1+I2 where I1 is equiv to L+D1 and I2 is D2
    #  - NFD(I1+I2) == L+D2+D1 because the canonical ordering wants D2 before D1
    #  - NFD(I1)+NFD(I2) == L+D1+D2 != NFD(I1+I2)
    # This also handles Input=L+D1+D2 going to output L+D2+D1
    # See test_utils.py for a real example
    for i, input_char in enumerate(inp):
        input_char_decomposed = ud.normalize(norm_form, input_char)
        for c in input_char_decomposed:
            found = False
            for j, (k, output_char) in enumerate(decomp_to_use):
                if output_char == c:
                    # Found c in the decomposed output, emit an arc and mark it consumed
                    indices.append((i, k))
                    decomp_to_use.pop(j)
                    found = True
                    break
                elif ud.category(output_char) != "Mn":
                    # No match, but got to a char that is not zero-width combining, so don't scan further
                    break
            if not found and decomp_to_use:  # pragma: no cover
                # c itself not found, fall back to the next unused character.
                # This should never happen, but it's here in case I missed something,
                # to make sure we don't end up in an infinite loop.
                k, _output_char = decomp_to_use.pop(0)
                indices.append((i, k))

    return decomposed, indices


def normalize_to_NFC_with_indices(
    inp: str, norm_form: str
) -> Tuple[str, List[Tuple[int, int]]]:
    """Normalize to NFC and return the indices mapping input to output characters"""
    assert norm_form in ("NFC", "NFKC")
    # Sadly mypy doesn't do narrowing to literals properly
    norm_form = cast(Literal["NFC", "NFKC"], norm_form)
    inp_nfc = ud.normalize(norm_form, inp)
    NFD_form = norm_form[:-1] + "D"  # NFC->NFD or NFKC->NFKD
    inp_nfd, indices_to_nfd = normalize_to_NFD_with_indices(inp, NFD_form)
    remapped_nfd, reverse_indices_to_nfc = normalize_to_NFD_with_indices(
        inp_nfc, NFD_form
    )
    assert inp_nfd == remapped_nfd
    indices_to_nfc = [(b, a) for a, b in reverse_indices_to_nfc]
    return inp_nfc, compose_indices(indices_to_nfd, indices_to_nfc)


def normalize_with_indices(
    inp: str, norm_form: str
) -> Tuple[str, List[Tuple[int, int]]]:
    """Normalize inp to the specified norm_form (NFC, NFD, NFKC, or NFKD)

    Returns:
        Tuple(normalized string, mapping indices)
    """
    if norm_form in ("NFC", "NFKC"):
        return normalize_to_NFC_with_indices(inp, norm_form)
    if norm_form in ("NFD", "NFKD"):
        return normalize_to_NFD_with_indices(inp, norm_form)
    if norm_form in ("none", None):
        return inp, [(i, i) for i in range(len(inp))]
    raise exceptions.InvalidNormalization(norm_form)


def unicode_escape(text):
    """Find any escaped characters and turn them into codepoints"""
    return re.sub(r"""\\(u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{6})""", escape_to_codepoint, text)


def escape_to_codepoint(match):
    """Turn escape into codepoint"""
    hex_codepoint = match.group(1)[1:]
    return chr(int(hex_codepoint, base=16))


EXPLICIT_INDEX_PATTERN = re.compile(r"{\d+}")


def strip_index_notation(string: str) -> str:
    """Return a string stripped of any explicit indices

    >>> strip_index_notation('test')
    'test'

    >>> strip_index_notation('t{0}e{2}st')
    'test'

    Args:
        string (str): a string that might have explicit indices

    Returns:
        str: a string without explicit indices
    """
    return re.sub(EXPLICIT_INDEX_PATTERN, "", string)


def create_fixed_width_lookbehind(pattern):
    """Turn all characters into fixed width lookbehinds"""
    return re.sub(
        re.compile(
            r"""
                # lookbehind
                (?<=\(?)
                # match any number of Unicode characters and diacritics, plus
                # square brackets, and backslash so patterns like \b can be used
                [\\\[\]\p{L}\p{M}|.:'^$]+
                # lookahead
                (?=\)?)
            """,
            re.U | re.VERBOSE,
        ),
        pattern_to_fixed_width_lookbehinds,
        pattern,
    )


def pattern_to_fixed_width_lookbehinds(match) -> str:
    """Python must have fixed-width lookbehinds."""
    pattern = match.group()
    pattern = sorted(pattern.split("|"), key=len, reverse=True)
    # in python ^ and $ have null length so must be ordered differently for proper
    # fixed-width lookbehinds
    null_length_characters = ["^", "$"]
    current_len = 0 if pattern[0] in null_length_characters else len(pattern[0])
    all_lookbehinds: List[Iterable[str]] = []
    current_list = []
    for item in pattern:
        item_length = 0 if item in null_length_characters else len(item)
        if item_length == current_len:
            current_list.append(item)
        else:
            current_len = item_length
            all_lookbehinds.append(current_list)
            current_list = [item]
        if pattern.index(item) == len(pattern) - 1:
            all_lookbehinds.append(current_list)
    all_lookbehinds_joined_by_length = [
        f"(?<={'|'.join(items)})" for items in all_lookbehinds
    ]
    return "(" + "|".join(all_lookbehinds_joined_by_length) + ")"


def load_from_workbook(language: str) -> List[dict]:
    """Parse mapping from Excel workbook"""
    from openpyxl import load_workbook  # Expensive import, do it only when needed

    try:
        work_book = load_workbook(language, read_only=True)
        work_sheet = work_book.active
        # Create wordlist
        mapping: List[dict] = []
        # Loop through rows in worksheet, create if statements for different columns
        # and append mappings to self.mapping.
        for row in work_sheet:  # type: ignore
            new_io = {"in": "", "out": "", "context_before": "", "context_after": ""}
            for cell in row:
                if cell.value is None:  # avoid tripping on empty cells
                    continue
                if cell.column in ["A", 1]:
                    new_io["in"] = str(cell.value)
                elif cell.column in ["B", 2]:
                    new_io["out"] = str(cell.value)
                elif cell.column in ["C", 3] and cell.value is not None:
                    new_io["context_before"] = str(cell.value)
                elif cell.column in ["D", 4] and cell.value is not None:
                    new_io["context_after"] = str(cell.value)
            mapping.append(new_io)
    finally:
        work_book.close()

    return mapping


def load_from_csv(language, delimiter=","):
    """Parse mapping from csv"""
    work_sheet = []
    with open(language, encoding="utf8") as f:
        reader = csv.reader(f, delimiter=delimiter)
        for line in reader:
            work_sheet.append(line)
    # Create wordlist
    mapping = []
    # Remove a well-formed header line
    if work_sheet and work_sheet[0][0:2] == ["in", "out"]:
        work_sheet = work_sheet[1:]
    # Loop through rows in worksheet, remove any stray BOMs
    # (zero-width non-breaking spaces), create if statements for
    # different columns and append mappings to self.mapping.
    remove_bom = str.maketrans("", "", "\ufeff")
    for entry in work_sheet:
        new_io = {"in": "", "out": "", "context_before": "", "context_after": ""}
        if len(entry) == 0:
            # Just ignore empty lines in the CSV file
            continue

        if len(entry) == 1:
            raise exceptions.MalformedMapping(
                'Entry {} in mapping {} has no "out" value.'.format(entry, language)
            )

        new_io["in"] = entry[0].translate(remove_bom)
        new_io["out"] = entry[1].translate(remove_bom)
        try:
            new_io["context_before"] = entry[2].translate(remove_bom)
        except IndexError:
            new_io["context_before"] = ""
        try:
            new_io["context_after"] = entry[3].translate(remove_bom)
        except IndexError:
            new_io["context_after"] = ""
        for k in new_io:
            if isinstance(new_io[k], (float, int)):
                new_io[k] = str(new_io[k])
        mapping.append(new_io)

    return mapping


def load_from_file(path: Union[Path, str]) -> list:
    """Helper method to load mapping from file."""
    path = str(path)
    if path.endswith("csv"):
        mapping = load_from_csv(path, ",")
    elif path.endswith("tsv"):
        mapping = load_from_csv(path, "\t")
    elif path.endswith("psv"):
        mapping = load_from_csv(path, "|")
    elif path.endswith("xlsx"):
        mapping = load_from_workbook(path)
    elif path.endswith("json"):
        with open(path, encoding="utf8") as f:
            mapping = json.load(f)
    else:
        raise exceptions.IncorrectFileType(
            f"File {path} is not a valid mapping filetype."
        )
    if not mapping:
        raise exceptions.MalformedMapping(
            f"Sorry, the file {path} does not contain any rules."
        )
    return mapping


def find_mapping_type(name):
    """Return the type of a mapping given its name"""
    if is_ipa(name):
        return "IPA"
    elif is_xsampa(name):
        return "XSAMPA"
    elif is_dummy(name):
        return "dummy"
    else:
        return "custom"


def escape_special_characters(to_escape: Union[Rule, Dict[str, str]]) -> Rule:
    if isinstance(to_escape, dict):
        to_escape = Rule(**to_escape)
    for key in ["rule_input", "context_before", "context_after"]:
        escaped = re.escape(getattr(to_escape, key))
        if getattr(to_escape, key) != escaped:
            LOGGER.debug(
                f"Escaped special characters in '{getattr(to_escape, key)}' with '{escaped}'. Set 'escape_special' "
                "to False in your Mapping configuration to disable this."
            )
        setattr(to_escape, key, escaped)
    return to_escape


def load_abbreviations_from_file(path: Union[str, Path]):
    """Helper method to load abbreviations from file."""
    path = str(path)
    if path.endswith("csv"):
        abbs = []
        with open(path, encoding="utf8") as f:
            reader = csv.reader(f, delimiter=",")
            abbs = flatten_abbreviations_format(reader)
    elif path.endswith("tsv"):
        abbs = []
        with open(path, encoding="utf8") as f:
            reader = csv.reader(f, delimiter="\t")
            abbs = flatten_abbreviations_format(reader)
    elif path.endswith("psv"):
        abbs = []
        with open(path, encoding="utf8") as f:
            reader = csv.reader(f, delimiter="|")
            abbs = flatten_abbreviations_format(reader)
    else:
        raise exceptions.IncorrectFileType(
            f"Sorry, abbreviations must be stored as CSV/TSV/PSV files. You provided the following: {path}"
        )
    return abbs


def get_alignment_input_string(alignment: str) -> str:
    """Parse one alignment of the format in *.aligned.txt and return just the input"""
    return "".join(
        [
            tok
            for mapping in alignment.split()
            for tok in mapping[: mapping.rindex("}")].split("|")
            if tok != "_"
        ]
    )


def get_alignment_sequence(alignment: str, delimiter="") -> List[Tuple[int, str]]:
    """Parse one alignment of the format in *.aligned.txt and return just the output seq.

    E.g.: a}ʌ b}b a}æ s|h}ʃ e|d}t (from cmudict.ipa.aligned.txt)
    means "abashed" is pronounced /ʌbæʃt/ with the grapheme-phoneme alignment being
    [(1, "ʌ"), (1, "b"), (1, "æ"), (2, "ʃ"), (2, "t")]

    Returns: the alignment as a List[Tuple[int, str]] where the int is the number
             of input characters consumed and the str are the output phoneme(s).
    """
    mappings: List[Tuple[int, str]] = []
    for mapping in alignment.split():
        idx = mapping.rindex("}")
        # Note that we care about *character* indices, so we join them together
        in_len = sum(len(tok) for tok in mapping[:idx].split("|") if tok != "_")
        out_seq = delimiter.join(
            tok for tok in mapping[idx + 1 :].split("|") if tok != "_"
        )
        # To save space, make the mappings flat and only store
        # the number of input characters rather than the characters themselves
        mappings.append((in_len, out_seq))
    return mappings


# The joiner between key and value must be 0 so that it sorts before all
# characters and thus won't break bisect_left()
_JOINER = "\0"
# For compacting a group of lexicon entries into one string.
# This just has to be somethign that does not occur in the lexicon data
_BLOCK_JOINER = "\1"


def find_alignment(alignments: List[str], word: str) -> List[Tuple[int, str]]:
    """Given a sorted list of (word, alignment), find word and return its parsed alignment.

    Algorithm: double bisect over blocks and then entries within blocks.
    """
    i = bisect_left(alignments, word)
    if i != len(alignments) and alignments[i].startswith(word + _JOINER):
        # Looking for the first entry of a block bisects to the correct block
        alignment_entry, _, _ = alignments[i].partition(_BLOCK_JOINER)
    elif i > 0:
        # Looking for the remaining entries of a block bisects one block too far:
        # bisect again within the previous block
        alignment_block = alignments[i - 1].split(_BLOCK_JOINER)
        j = bisect_left(alignment_block, word)
        if j != len(alignment_block):
            alignment_entry = alignment_block[j]
        else:
            return []  # word not found: would have been between this and next block
    else:
        return []  # word not found: would have been before the first block

    k, _, v = alignment_entry.partition(_JOINER)
    if k == word:
        return get_alignment_sequence(v)  # word found
    else:
        return []  # word not found: key in bisected location does not match word


def compact_alignments(alignments: Sequence[str]) -> List[str]:
    """Memory footprint optimization: compact the list of alignments into blocks.

    Each Python string has a significant overhead: grouping them into blocks of 16
    saves 15MB of RAM for the cmudict English lexicon, at no significant speed cost.
    """
    _BLOCK_SIZE = 16
    return [
        _BLOCK_JOINER.join(alignments[i : i + _BLOCK_SIZE])
        for i in range(0, len(alignments), _BLOCK_SIZE)
    ]


def load_alignments_from_file(path, delimiter="") -> List[str]:
    """Load alignments in Phonetisaurus default format.

    Returns a mapping of input words to output alignments used to
    create a lexicon mapping.  These are of the form (length,
    outputs, length, outputs, ...) - that is, a sequence of pairs
    specifying how much of the input to consume and what it maps
    to.  This particular format is used to avoid redundancy with
    the keys in the dictionary.
    """
    LOGGER.info("Loading alignments from %s", path)
    alignments = []
    with open(path, encoding="utf8") as f:
        for spam in f:
            spam = spam.strip()
            if not spam:
                continue
            word = get_alignment_input_string(spam)
            alignments.append(word + _JOINER + spam)
    return compact_alignments(sorted(alignments))


def is_ipa(lang: str) -> bool:
    pattern = re.compile("[-_]?ipa$")
    return bool(re.search(pattern, lang))


def is_xsampa(lang: str) -> bool:
    pattern = re.compile("[-_]?x(-?)sampa$")
    return bool(re.search(pattern, lang))


def is_dummy(lang: str) -> bool:
    return lang == "dummy"


class IndentDumper(yaml.Dumper):
    def increase_indent(self, flow=False, indentless=False):
        return super().increase_indent(flow, False)

    def ignore_aliases(self, *_args):
        return True


def merge_same_type_tokens(tokens: List[Token]) -> List[Token]:
    """Merge tokens that have the same type.  Destroys tokens in the process.

    >>> merge_same_type_tokens([Token("test", True), Token("b", True), Token(":", False), Token(",", False)])
    [Token(text='testb', is_word=True), Token(text=':,', is_word=False)]
    >>> merge_same_type_tokens([])
    []
    """
    if not tokens:
        return []
    merged_tokens = [tokens[0]]
    for token in tokens[1:]:
        if token.is_word == merged_tokens[-1].is_word:
            merged_tokens[-1].text += token.text
        else:
            merged_tokens.append(token)
    return merged_tokens


def split_non_word_tokens(tokens: List[Token]) -> List[Token]:
    """Split non-word units into characters. Reuses the word tokens.

    Generates a maximum of 5 units per non-word token: if the input token is
    more than 5 non-word characters, the output will be the first two
    individually, the middle as a block, and the last two individually, because
    lexicon-based tokenization does not need more granularity than that.
    This prevents degenerate input like a large number of consecutive punctuation
    marks from taking quadratic time in lexicon-based tokenization.

    >>> split_non_word_tokens([Token("test", True), Token(":,- ", False), Token("", False)])
    [Token(text='test', is_word=True), Token(text=':', is_word=False), Token(text=',', is_word=False), Token(text='-', is_word=False), Token(text=' ', is_word=False)]
    >>> split_non_word_tokens([])
    []
    >>> split_non_word_tokens([Token(".,.,.,.", False)])
    [Token(text='.', is_word=False), Token(text=',', is_word=False), Token(text='.,.', is_word=False), Token(text=',', is_word=False), Token(text='.', is_word=False)]
    """
    new_tokens = []
    for token in tokens:
        if not token.is_word:
            text = token.text
            if len(text) > 5:
                new_tokens.append(Token(text[0], False))
                new_tokens.append(Token(text[1], False))
                new_tokens.append(Token(text[2:-2], False))
                new_tokens.append(Token(text[-2], False))
                new_tokens.append(Token(text[-1], False))
            else:
                new_tokens.extend([Token(char, False) for char in text])
        else:
            new_tokens.append(token)
    return new_tokens


def merge_non_word_tokens(tokens: List[Token]) -> List[Token]:
    """Merge consecutive non-word units into a single token. Destroys tokens in the process.

    >>> merge_non_word_tokens([Token("test", True), Token(":", False), Token(",", False)])
    [Token(text='test', is_word=True), Token(text=':,', is_word=False)]
    >>> merge_non_word_tokens([])
    []
    """
    if not tokens:
        return tokens
    merged_tokens = [tokens[0]]
    for token in tokens[1:]:
        if not token.is_word and not merged_tokens[-1].is_word:
            merged_tokens[-1].text += token.text
        else:
            merged_tokens.append(token)
    return merged_tokens


CATEGORIES = {
    "Cc": "other",  # Other, Control
    "Cf": "other",  # Other, Format
    "Cn": "other",  # Other, Not Assigned (no characters in the file have this property)
    "Co": "letter",  # Other, Private Use
    "Cs": "other",  # Other, Surrogate
    "LC": "letter",  # Letter, Cased
    "Ll": "letter",  # Letter, Lowercase
    "Lm": "letter",  # Letter, Modifier
    "Lo": "letter",  # Letter, Other
    "Lt": "letter",  # Letter, Titlecase
    "Lu": "letter",  # Letter, Uppercase
    "Mc": "diacritic",  # Mark, Spacing Combining
    "Me": "diacritic",  # Mark, Enclosing
    "Mn": "diacritic",  # Mark, Nonspacing
    "Nd": "number",  # Number, Decimal Digit
    "Nl": "number",  # Number, Letter
    "No": "number",  # Number, Other
    "Pc": "punctuation",  # Punctuation, Connector
    "Pd": "punctuation",  # Punctuation, Dash
    "Pe": "punctuation",  # Punctuation, Close
    "Pf": "punctuation",  # Punctuation, Final quote (may behave like Ps or Pe depending on usage)
    "Pi": "punctuation",  # Punctuation, Initial quote (may behave like Ps or Pe depending on usage)
    "Po": "punctuation",  # Punctuation, Other
    "Ps": "punctuation",  # Punctuation, Open
    "Sc": "symbol",  # Symbol, Currency
    "Sk": "symbol",  # Symbol, Modifier
    "Sm": "symbol",  # Symbol, Math
    "So": "symbol",  # Symbol, Other
    "Zl": "whitespace",  # Separator, Line
    "Zp": "whitespace",  # Separator, Paragraph
    "Zs": "whitespace",  # Separator, Space
}


def get_unicode_category(c):
    """Maps a character to one of [ "letter", "number", "diacritic", "punctuation",
    "symbol", "whitespace", "other"]"""
    cat = ud.category(c)
    assert cat in CATEGORIES
    return CATEGORIES[cat]


class CompactJSONMappingEncoder(json.JSONEncoder):
    """A JSON Encoder that puts individual rules on a single line.

    This way, it's easy to look at a generated mapping visually.

    This code is adapted from https://stackoverflow.com/questions/16264515/json-dumps-custom-formatting
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.indentation_level = 0

    def encode(self, obj):
        if isinstance(obj, dict) and "in" in obj and "out" in obj:
            # Compact, single-line output for the individual rules in the mapping
            ordered_rule = {"in": obj["in"], "out": obj["out"], **obj}
            assert obj == ordered_rule
            return json.dumps(ordered_rule, ensure_ascii=self.ensure_ascii)
        elif isinstance(obj, (list, tuple)):
            self.indentation_level += 1
            output = [self.indent_str + self.encode(el) for el in obj]
            self.indentation_level -= 1
            return "[\n" + ",\n".join(output) + "\n" + self.indent_str + "]"
        elif isinstance(obj, dict):
            self.indentation_level += 1
            output = [
                self.indent_str + f"{json.dumps(k)}: {self.encode(v)}"
                for k, v in obj.items()
            ]
            self.indentation_level -= 1
            return "{\n" + ",\n".join(output) + "\n" + self.indent_str + "}"
        else:
            return json.dumps(obj, ensure_ascii=self.ensure_ascii)

    @property
    def indent_str(self) -> str:
        return " " * self.indentation_level * self.indent  # type: ignore[operator]

    def iterencode(self, o, *_args, **_kwargs):
        return self.encode(o)


class MAPPING_TYPE(str, Enum):
    mapping = "mapping"
    unidecode = "unidecode"
    lexicon = "lexicon"
    neural = "neural"


class NEURAL_MODEL_HANDLER(str, Enum):
    deepphonemizer = "deepphonemizer"


class NORM_FORM_ENUM(str, Enum):
    NFC = "NFC"
    NFD = "NFD"
    NKFC = "NKFC"
    NKFD = "NKFD"
    none = "none"


class RULE_ORDERING_ENUM(str, Enum):
    as_written = "as-written"
    apply_longest_first = "apply-longest-first"


class _MappingModelDefinition(BaseModel):
    processed: bool = Field(False, json_schema_extra={"hidden": True})

    parent_dir: Optional[DirectoryPath] = None
    """Optionally resolve all paths to a parent directory"""

    id: Optional[str] = None
    """A unique ID for the mapping"""

    in_lang: str = "standalone"
    """The input language ID"""

    out_lang: str = "standalone"
    """The output language ID"""

    language_name: Optional[str] = None
    """The name of the language"""

    display_name: Optional[str] = None
    """The display name of the mapping"""

    as_is: Optional[bool] = None
    """Deprecated: Please use rule_ordering='as_written' """

    case_sensitive: bool = True
    """When false, lowercase all rules and conversion input"""

    case_equivalencies: dict = {}
    """List of case equivalencies for preserve_case that are not already in the Unicode standard"""

    preserve_case: bool = False
    """Preserve source case in output (requires case_sensitive=False)"""

    escape_special: bool = False
    """Escape special characters in rules"""

    norm_form: NORM_FORM_ENUM = NORM_FORM_ENUM.NFD
    """Normalization standard to follow"""

    out_delimiter: str = ""
    """Separate output transformations with a delimiter"""

    reverse: bool = False
    """Reverse all mappings."""

    rule_ordering: RULE_ORDERING_ENUM = RULE_ORDERING_ENUM.as_written
    """ Affects in what order the rules are applied.

        If set to ``"as-written"``, rules are applied from top-to-bottom in the order that they
        are written in the source file
        (previously this was accomplished with ``as_is=True``).

        If set to ``"apply-longest-first"``, rules are first sorted such that rules with the longest
        input are applied first. Sorting the rules like this prevents shorter rules
        from taking part in feeding relations
        (previously this was accomplished with ``as_is=False``)
    """

    prevent_feeding: bool = False
    """Converts each rule into an intermediary form in the Unicode PUA"""

    type: Optional[MAPPING_TYPE] = None
    """Type of mapping, either "mapping" (rules), "unidecode" (magical Unicode guessing),
        "lexicon" (lookup in an aligned lexicon), or "neural" (3rd party neural model, only available for some languages)."""

    alignments: List[str] = []
    """The alignments for a lexicon mapping"""

    alignments_path: Optional[Path] = None
    """A path specifying a file from which to load alignments when type = 'lexicon'"""

    authors: Optional[List[str]] = None
    """A list of authors responsible for the mapping."""

    abbreviations: Dict[str, List[str]] = {}
    """A list of 'abbreviations' for your mappings.
    Please see https://blog.mothertongues.org/g2p-advanced-mappings/ for more information.
    """

    abbreviations_path: Optional[Path] = None
    """A path to an 'abbreviations' file"""

    rules: List[Rule] = []
    """A list of Rules"""

    rules_path: Optional[Path] = None
    """A path to a file of a list of rules"""

    hf_model_identifier: Optional[str] = None
    """The identifier for a model hosted on HuggingFace, if applicable"""

    hf_model_filepath: Optional[str] = None
    """The path to the specific model on HuggingFace if downloading the entire repo is not needed"""

    neural_model_handler: Optional[NEURAL_MODEL_HANDLER] = None

    neural_kwargs: Optional[dict] = None

    model_config = ConfigDict(str_strip_whitespace=False, extra="allow")

    @model_validator(mode="after")
    def check_mapping_types(self) -> "_MappingModelDefinition":
        if (
            (self.type == MAPPING_TYPE.mapping or self.type is None)
            and not self.rules
            and self.rules_path is None
        ):
            LOGGER.warning(
                "Empty mapping: specify some rules or a path to a file containing rules."
            )
        if (
            (self.type == MAPPING_TYPE.lexicon)
            and not self.alignments
            and self.alignments_path is None
        ):
            raise exceptions.MalformedMapping(
                "Lexicon mappings must also provide alignments"
            )
        return self

    @field_serializer("rules_path", "abbreviations_path", "alignments_path")
    def serialize_paths(self, path: Path):
        return path.name if path else path

    @field_serializer("rules")
    def serialize_mapping(self, rules: List[Rule]):
        return [rule.export_to_dict() for rule in rules]

    @field_validator("norm_form", mode="before")
    @classmethod
    def validate_norm_form(cls, v):
        if not v or v is None:
            v = "none"
        return v

    @field_validator("case_equivalencies", mode="before")
    @classmethod
    def validate_case_equivalencies(cls, v):
        if not v or v is None:
            v = {}
        for lower_case, upper_case in v.items():
            if len(lower_case) != len(upper_case):
                raise exceptions.MalformedMapping(
                    f"Sorry, the case equivalency between {lower_case} and {upper_case} is not valid because it "
                    "is not the same length, please write rules such that any case equivalent is of equal length."
                )
        return v

    @model_validator(mode="after")
    def validate_preserve_case(self):
        """preserve_case=True requires case_sensitive=False"""
        if self.preserve_case and self.case_sensitive:
            raise exceptions.MalformedMapping(
                "Sorry, preserve_case=True requires case_sensitive=False."
            )
        return self

    @field_validator(
        "rules_path", "abbreviations_path", "alignments_path", mode="before"
    )
    @classmethod
    def add_parent_dir(cls, value: Any, info: ValidationInfo):
        """If there is a parent directory, prepend it to all path fields."""
        if isinstance(value, (str, Path)):
            if info.data.get("parent_dir", None):
                value = Path(info.data["parent_dir"]) / value
        return value

    @model_validator(mode="before")
    @classmethod
    def create_language_name(cls, data: Any):
        """When missing, default language_name to in_lang"""
        if isinstance(data, dict) and data.get("language_name", None) is None:
            data["language_name"] = data.get("in_lang", None)
        return data

    @model_validator(mode="before")
    @classmethod
    def create_display_name(cls, data: Any):
        """When missing, create a default display_name from in_lang and out_lang"""
        if isinstance(data, dict) and data.get("display_name", None) is None:
            in_lang = data.get("in_lang", "")
            out_lang = data.get("out_lang", "")
            data["display_name"] = (
                f"{in_lang} {find_mapping_type(in_lang)} to "
                f"{out_lang} {find_mapping_type(out_lang)}"
            )
        return data



import unittest
from contextlib import redirect_stderr
from io import StringIO

from fastapi.testclient import TestClient

from g2p.api_v2 import api

API_CLIENT = TestClient(api)


class TestAPIV2(unittest.TestCase):
    def test_langs(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.get("/langs")
        self.assertEqual(response.status_code, 200)
        codes = {x["code"] for x in response.json()}
        self.assertTrue("fin" in codes)
        self.assertTrue("atj" in codes)
        self.assertFalse("generated" in codes)
        self.assertFalse("atj-ipa" in codes)
        names = {x["name"] for x in response.json()}
        self.assertTrue("Finnish" in names)

    def test_langs_allcodes(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.get("/nodes")
        self.assertEqual(response.status_code, 200)
        codes = {x["code"] for x in response.json()}
        self.assertTrue("eng-arpabet" in codes)
        self.assertTrue("eng-ipa" in codes)
        self.assertTrue("atj" in codes)
        self.assertFalse("generated" in codes)

    def test_outputs_for(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.get("/outputs_for/fin")
        self.assertEqual(response.status_code, 200)
        self.assertTrue("eng-arpabet" in response.json())
        self.assertTrue("eng-ipa" in response.json())

    def test_inputs_for(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.get("/inputs_for/eng-arpabet")
        self.assertEqual(response.status_code, 200)
        self.assertTrue("fin" in response.json())
        self.assertTrue("eng-ipa" in response.json())

    def test_convert(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.post(
                "/convert",
                json={
                    "in_lang": "fin",
                    "out_lang": "eng-arpabet",
                    "text": "hyvää yötä",
                },
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            [
                {
                    "conversions": [
                        {
                            "in_lang": "eng-ipa",
                            "out_lang": "eng-arpabet",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["h", "HH "],
                                ["u", "UW "],
                                ["w", "W "],
                                ["æ", "AE "],
                            ],
                        },
                        {
                            "in_lang": "fin-ipa",
                            "out_lang": "eng-ipa",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["h", "h"],
                                ["y", "u"],
                                ["ʋ", "w"],
                                ["æː", "æ"],
                            ],
                        },
                        {
                            "in_lang": "fin",
                            "out_lang": "fin-ipa",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["h", "h"],
                                ["y", "y"],
                                ["v", "ʋ"],
                                ["ä", "æ"],
                                ["ä", "ː"],
                            ],
                        },
                    ]
                },
                {
                    "conversions": [
                        {
                            "in_lang": None,
                            "out_lang": None,
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [[" ", " "]],
                        }
                    ]
                },
                {
                    "conversions": [
                        {
                            "in_lang": "eng-ipa",
                            "out_lang": "eng-arpabet",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["u", "UW "],
                                ["ə", "AH "],
                                ["t", "T "],
                                ["æ", "AE "],
                            ],
                        },
                        {
                            "in_lang": "fin-ipa",
                            "out_lang": "eng-ipa",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["y", "u"],
                                ["ø", "ə"],
                                ["t", "t"],
                                ["æ", "æ"],
                            ],
                        },
                        {
                            "in_lang": "fin",
                            "out_lang": "fin-ipa",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["y", "y"],
                                ["ö", "ø"],
                                ["t", "t"],
                                ["ä", "æ"],
                            ],
                        },
                    ]
                },
            ],
        )

    def test_convert_composed(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.post(
                "/convert",
                json={
                    "in_lang": "fin",
                    "out_lang": "eng-arpabet",
                    "text": "hyvää yötä",
                    "compose_from": "fin",
                },
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            [
                {
                    "conversions": [
                        {
                            "in_lang": "fin",
                            "out_lang": "eng-arpabet",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["h", "HH "],
                                ["y", "UW "],
                                ["v", "W "],
                                ["ää", "AE "],
                            ],
                        },
                    ],
                },
                {
                    "conversions": [
                        {
                            "in_lang": None,
                            "out_lang": None,
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [[" ", " "]],
                        }
                    ]
                },
                {
                    "conversions": [
                        {
                            "in_lang": "fin",
                            "out_lang": "eng-arpabet",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["y", "UW "],
                                ["ö", "AH "],
                                ["t", "T "],
                                ["ä", "AE "],
                            ],
                        },
                    ]
                },
            ],
        )

    def test_convert_compose_from(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.post(
                "/convert",
                json={
                    "in_lang": "fin",
                    "out_lang": "eng-arpabet",
                    "text": "hyvää yötä",
                    "compose_from": "fin-ipa",
                },
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            [
                {
                    "conversions": [
                        {
                            "in_lang": "fin-ipa",
                            "out_lang": "eng-arpabet",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["h", "HH "],
                                ["y", "UW "],
                                ["ʋ", "W "],
                                ["æː", "AE "],
                            ],
                        },
                        {
                            "in_lang": "fin",
                            "out_lang": "fin-ipa",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["h", "h"],
                                ["y", "y"],
                                ["v", "ʋ"],
                                ["ä", "æ"],
                                ["ä", "ː"],
                            ],
                        },
                    ],
                },
                {
                    "conversions": [
                        {
                            "in_lang": None,
                            "out_lang": None,
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [[" ", " "]],
                        }
                    ]
                },
                {
                    "conversions": [
                        {
                            "in_lang": "fin-ipa",
                            "out_lang": "eng-arpabet",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["y", "UW "],
                                ["ø", "AH "],
                                ["t", "T "],
                                ["æ", "AE "],
                            ],
                        },
                        {
                            "in_lang": "fin",
                            "out_lang": "fin-ipa",
                            "input_nodes": None,
                            "output_nodes": None,
                            "alignments": None,
                            "substring_alignments": [
                                ["y", "y"],
                                ["ö", "ø"],
                                ["t", "t"],
                                ["ä", "æ"],
                            ],
                        },
                    ]
                },
            ],
        )

    def test_convert_surrogates(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.post(
                "/convert",
                json={
                    "in_lang": "eng-ipa",
                    "out_lang": "eng-arpabet",
                    "text": "hi🙂hi",
                    "indices": True,
                    "tokenize": False,
                },
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            [
                {
                    "conversions": [
                        {
                            "in_lang": "eng-ipa",
                            "out_lang": "eng-arpabet",
                            "input_nodes": ["h", "i", "🙂", "h", "i"],
                            "output_nodes": [
                                "H",
                                "H",
                                " ",
                                "I",
                                "Y",
                                " ",
                                "🙂",
                                "H",
                                "H",
                                " ",
                                "I",
                                "Y",
                                " ",
                            ],
                            "alignments": [
                                [0, 0],
                                [0, 1],
                                [0, 2],
                                [1, 3],
                                [1, 4],
                                [1, 5],
                                [2, 6],
                                [3, 7],
                                [3, 8],
                                [3, 9],
                                [4, 10],
                                [4, 11],
                                [4, 12],
                            ],
                            "substring_alignments": [
                                ["h", "HH "],
                                ["i", "IY "],
                                ["🙂", "🙂"],
                                ["h", "HH "],
                                ["i", "IY "],
                            ],
                        }
                    ]
                }
            ],
        )

    def test_convert_no_path(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.post(
                "/convert",
                json={
                    "in_lang": "fin",
                    "out_lang": "fra",
                    "text": "hyvää yötä",
                },
            )
        self.assertEqual(response.status_code, 400)

    def test_convert_invalid(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.post(
                "/convert",
                json={
                    "in_lang": "Finnish",
                    "out_lang": "eng-arpabet",
                    "text": "hyvää yötä",
                },
            )
        self.assertEqual(response.status_code, 422)
        self.assertIn("Input should be", response.json()["detail"][0]["msg"])

    def test_path(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.get("/path/fin/eng-arpabet")
        self.assertEqual(response.status_code, 200)
        path = response.json()
        self.assertEqual(path, ["fin", "fin-ipa", "eng-ipa", "eng-arpabet"])

    def test_no_path(self):
        with redirect_stderr(StringIO()):
            response = API_CLIENT.get("/path/fin/fra")
        self.assertEqual(response.status_code, 400)
        self.assertIn("No path", response.json()["detail"])


if __name__ == "__main__":
    unittest.main()



"""
The g2p Studio web app.

You can run the app (and API) for development purposes on any platform with:
    pip install uvicorn
    uvicorn g2p.app:APP --reload --port 5000
- The --reload switch will watch for changes under the directory where it's
  running and reload the code whenever it changes.

You can also spin up the app server grade (on Linux, not Windows) with gunicorn:
    gunicorn -w 4 -k uvicorn.workers.UvicornWorker g2p.app:APP --port 5000

Once spun up, the application will be visible at
http://localhost:5000/ and the API at http://localhost:5000/api/v1/docs

"""

import os
from pathlib import Path
from typing import Dict, List, Union

import socketio  # type: ignore
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import HTMLResponse, RedirectResponse
from starlette.routing import Mount, Route
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates

import g2p
from g2p import make_g2p, make_tokenizer
from g2p.api import api as api_v1
from g2p.api_v2 import api as api_v2
from g2p.log import LOGGER
from g2p.mappings import Mapping, Rule
from g2p.mappings.langs import LANGS_NETWORK
from g2p.mappings.utils import (
    _MappingModelDefinition,
    expand_abbreviations_format,
    flatten_abbreviations_format,
)
from g2p.transducer import (
    BaseTransducer,
    CompositeTransducer,
    CompositeTransductionGraph,
    TokenizingTransducer,
    Transducer,
    TransductionGraph,
)

DEFAULT_N = 10

BASE_DIR = Path(g2p.__file__).parent

TEMPLATES = Jinja2Templates(directory=BASE_DIR / "templates")
server_args: Dict[str, Union[bool, str]] = {"async_mode": "asgi"}
if os.getenv("G2P_STUDIO_DEBUG"):
    server_args["logger"] = server_args["engineio_logger"] = True
SIO = socketio.AsyncServer(**server_args)
SIO_APP = socketio.ASGIApp(socketio_server=SIO, socketio_path="/ws/socket.io")


async def home(request: Request) -> HTMLResponse:
    """Return homepage of g2p studio"""
    return TEMPLATES.TemplateResponse(request, "index.html")


async def redirect_to_docs(request: Request) -> RedirectResponse:
    """Redirect to v1 API docs for backward compatibility"""
    return RedirectResponse(url="/api/v1/docs/")


async def redirect_to_openapi(request: Request) -> RedirectResponse:
    """Redirect to v1 API JSON for backward compatibility"""
    return RedirectResponse(url="/api/v1/openapi.json")


APP = Starlette(
    debug=True,
    routes=[
        Route("/", home),
        Mount("/ws", SIO_APP),
        Mount("/api/v1", api_v1),
        Mount("/api/v2", api_v2),
        Route("/static/swagger.json", redirect_to_openapi),
        Mount("/static", StaticFiles(directory=BASE_DIR / "static"), name="static"),
        Route("/docs", redirect_to_docs),
    ],
)


def shade_colour(colour, percent, r=0, g=0, b=0):
    R = hex(min(255, int((int(colour[1:3], 16) * (100 + percent + r) / 100)))).lstrip(
        "0x"
    )
    G = hex(min(255, int((int(colour[3:5], 16) * (100 + percent + g) / 100)))).lstrip(
        "0x"
    )
    B = hex(min(255, int((int(colour[5:], 16) * (100 + percent + b) / 100)))).lstrip(
        "0x"
    )
    return "#" + str(R) + str(G) + str(B)


def contrasting_text_color(hex_str):
    (R, G, B) = (hex_str[1:3], hex_str[3:5], hex_str[5:])
    return (
        "#000"
        if 1 - (int(R, 16) * 0.299 + int(G, 16) * 0.587 + int(B, 16) * 0.114) / 255
        < 0.5
        else "#fff"
    )


def return_echart_data(tg: Union[CompositeTransductionGraph, TransductionGraph]):
    x = 100
    diff = 200
    nodes = []
    edges = []
    index_offset = 0
    colour = "#222222"
    if isinstance(tg, CompositeTransductionGraph):
        steps = len(tg.tiers)
    else:
        steps = 1
    for ind, tier in enumerate(tg.tiers):
        if ind == 0:
            symbol_size = round(min(300 / max(1, len(tier.input_string)), 40), 2)
            input_x = x + (ind * diff)
            input_y = 300
            x += diff
            inputs = [
                {
                    "name": f"{x}",
                    "id": f"(in{ind + index_offset}-in{i + index_offset})",
                    "x": input_x,
                    "y": input_y + (i * 50),
                    "symbolSize": symbol_size,
                    "label": {"color": contrasting_text_color(colour)},
                    "itemStyle": {
                        "color": colour,
                        "borderColor": contrasting_text_color(colour),
                    },
                }
                for i, x in enumerate(tier.input_string)
            ]
            nodes += inputs
        edges += [
            {
                "source": x[0] + index_offset,
                "target": x[1] + len(tier.input_string) + index_offset,
            }
            for x in tier.edges
            if x[1] is not None
        ]
        index_offset += len(tier.input_string)
        symbol_size = round(min(300 / max(1, len(tier.output_string)), 40), 2)
        colour = shade_colour(colour, (1 / steps) * 350, g=50, b=20)
        output_x = x + (ind * diff)
        output_y = 300
        outputs = [
            {
                "name": f"{x}",
                "id": f"({ind + index_offset}-{i + index_offset})",
                "x": output_x,
                "y": output_y + (i * 50),
                "symbolSize": symbol_size,
                "label": {"color": contrasting_text_color(colour)},
                "itemStyle": {
                    "color": colour,
                    "borderColor": contrasting_text_color(colour),
                },
            }
            for i, x in enumerate(tier.output_string)
        ]
        nodes += outputs
    return nodes, edges


@SIO.on("conversion event", namespace="/convert")  # type: ignore
async def convert(sid, message):
    """Convert input text and return output"""
    transducers = []
    LOGGER.debug("/convert: %s", message)
    transducer: BaseTransducer
    for mapping in message["data"]["mappings"]:
        mapping_args = {**mapping["kwargs"]}
        mapping_args["abbreviations"] = flatten_abbreviations_format(
            mapping["abbreviations"]
        )
        if mapping_args["type"] == "lexicon":
            lexicon = Mapping.find_mapping(
                mapping_args["in_lang"], mapping_args["out_lang"]
            )
            mapping_args["alignments"] = lexicon.alignments
        else:
            mapping_args["rules"] = mapping["rules"]
        try:
            mappings_obj = Mapping(**mapping_args)
            transducer = Transducer(mappings_obj)
            transducers.append(transducer)
        except Exception as e:
            LOGGER.warning(
                "Skipping invalid mapping %s->%s:\n%s",
                mapping_args["in_lang"],
                mapping_args["out_lang"],
                e,
            )
    if len(transducers) == 0:
        # This happens when we switch between output langs in g2p-studio
        await SIO.emit(
            "conversion response",
            {"output_string": message["data"]["input_string"]},
            sid,
            namespace="/convert",
        )
        return
    transducer = CompositeTransducer(transducers)
    try:
        in_lang = message["data"]["mappings"][0]["kwargs"]["in_lang"]
        tokenizer = make_tokenizer(in_lang)
        transducer = TokenizingTransducer(transducer, tokenizer)
    except Exception:
        # We tokenize by default when it works (see try block) but just skip
        # tokenization in case of problem: it's not a bug, tokenization is optional.
        pass
    if message["data"]["index"]:
        tg = transducer(message["data"]["input_string"])
        data, links = return_echart_data(tg)
        await SIO.emit(  # type: ignore
            "conversion response",
            {
                "output_string": tg.output_string,
                "index_data": data,
                "index_links": links,
            },
            sid,
            namespace="/convert",
        )
    else:
        output_string = transducer(message["data"]["input_string"]).output_string
        await SIO.emit(  # type: ignore
            "conversion response",
            {"output_string": output_string},
            sid,
            namespace="/convert",
        )


@SIO.on("table event", namespace="/table")  # type: ignore
async def change_table(sid, message) -> None:
    """Change the lookup table"""
    LOGGER.debug("/table: %s", message)
    if "in_lang" not in message or "out_lang" not in message:
        await SIO.emit(
            "table response",
            [],
            sid,
            namespace="/table",
        )
    elif message["in_lang"] == "custom" or message["out_lang"] == "custom":
        # These are only used to generate JSON to send to the client,
        # so it's safe to create a list of references to the same thing.
        mapping_dicts = [
            {"in": "", "out": "", "context_before": "", "context_after": ""}
        ] * DEFAULT_N
        abbs = [[""] * 6] * DEFAULT_N

        kwargs = _MappingModelDefinition(
            language_name="Custom",
            display_name="Custom",
            in_lang="custom",
            out_lang="custom",
            type="mapping",
            norm_form="NFC",
            # Put something here to silence a warning
            rules=[Rule(rule_input="a", rule_output="a")],
        ).model_dump()
        kwargs["rules"] = []
        # Remove the bogus rule we used to silence the validator
        kwargs["include"] = False
        await SIO.emit(
            "table response",
            [
                {
                    "mappings": mapping_dicts,
                    "abbs": abbs,
                    "kwargs": kwargs,
                }
            ],
            sid,
            namespace="/table",
        )
    else:
        path = LANGS_NETWORK.shortest_path(message["in_lang"], message["out_lang"])
        mappings: List[Mapping] = []
        for lang1, lang2 in zip(path[:-1], path[1:]):
            transducer = make_g2p(lang1, lang2, tokenize=False)
            mappings.append(transducer.mapping)  # type: ignore
        await SIO.emit(
            "table response",
            [
                {
                    "mappings": mapping.plain_mapping(),
                    "abbs": expand_abbreviations_format(mapping.abbreviations),
                    "kwargs": mapping.model_dump(exclude={"alignments"}),
                }
                for mapping in mappings
            ],
            sid,
            namespace="/table",
        )


@SIO.on("connect", namespace="/connect")  # type: ignore
async def test_connect(sid, message):
    """Let client know disconnected"""
    await SIO.emit(  # type: ignore
        "connection response", {"data": "Connected"}, sid, namespace="/connect"
    )


@SIO.on("disconnect", namespace="/connect")  # type: ignore
async def test_disconnect(sid):
    """Let client know disconnected"""
    await SIO.emit(  # type: ignore
        "connection response", {"data": "Disconnected"}, sid, namespace="/connect"
    )



"""
This module contains the Transducer and CompositeTransducer classes
which are responsible for performing transductions in the g2p library.
"""

import copy
import re
import unicodedata
from collections import OrderedDict, defaultdict
from functools import partial
from typing import DefaultDict, Dict, List, Optional, Set, Tuple, Union

import text_unidecode  # type: ignore

from g2p.log import LOGGER
from g2p.mappings import MAPPING_TYPE, Mapping
from g2p.mappings.langs.utils import is_arpabet, is_panphon
from g2p.mappings.utils import (
    NEURAL_MODEL_HANDLER,
    Rule,
    compose_indices,
    deep_phonemizer_handler,
    download_huggingface_model,
    find_alignment,
    is_ipa,
    normalize,
    normalize_with_indices,
    strip_index_notation,
    unicode_escape,
)
from g2p.shared_types import BaseTokenizer, BaseTransducer, BaseTransductionGraph

# An Index is typed as follows:
# {input_index: int, {'input_string': str, 'output': {output_index: int, str}}}
# Example:
# {0: {'input_string': 'h', 'output': {0: 'ʔ'}}}
Index = Dict

# A ChangeLog is a list of changes (List[int])
# The first item (int) in a change is the index of where the change occurs, and
# the second item (int) is the change offset
# Example:
# an insertion of length 1 at index 0 followed by a deletion of length one at index 2
# [[0,1],[2,-1]]
ChangeLog = List[List[int]]

UNIDECODE_SPECIALS = ["@", "?", "'", ",", ":"]


def normalize_edges(
    edges: List[Tuple[int, Optional[int]]],
) -> List[Tuple[int, Optional[int]]]:
    """Normalize and sort a list of edges.

    Specifically, this:

    - Removes any non-deletion edges with the same input index as a deletion
    - Resolves all deletions to map the outputs to the previous index if possible,
      the following index if possible, or otherwise None (in practice this means
      that None only occurs if the output is empty)
    - Sorts edges based on the input and suppresses duplicates
    """
    to_remove: Set[int] = set()
    for i, edge in enumerate(edges):
        if edge[1] is None:
            for j, other_edge in enumerate(edges):
                if other_edge != edge and other_edge[0] == edge[0]:
                    to_remove.add(j)
    edges = [edge for i, edge in enumerate(edges) if i not in to_remove]
    # sort based on inputs
    edges.sort(key=lambda x: x[0])
    for i, edge in enumerate(edges):
        if edge[1] is None:
            # if previous exists, use that, otherwise use following, otherwise None
            previous = [x for x in edges[:i] if x[1] is not None]
            # Note: if len(edges) == 1, edges[1:] is empty, not an IndexError
            following = [x for x in edges[i + 1 :] if x[1] is not None]
            if previous:
                edges[i] = (edge[0], previous[-1][1])
            elif following:
                edges[i] = (edge[0], following[0][1])
    # uniquify preserving order
    return list(OrderedDict.fromkeys((i, j) for i, j in edges))


class TransductionGraph(BaseTransductionGraph):
    """This is the object returned after performing a transduction using a Transducer.

    Each TransductionGraph must be initialized with an input string.
    """

    def __init__(self, input_string: str):
        # Plain strings
        self._input_string = input_string
        self._output_string = input_string
        # Nodes
        self._input_nodes: List[Tuple[int, str]] = list(enumerate(input_string))
        self._output_nodes: List[Tuple[int, str]] = list(enumerate(input_string))
        # Edges
        self._edges: List[Tuple[int, Optional[int]]] = [
            (i, i) for i in range(len(input_string))
        ]
        # Debugger
        self._debugger = []  # type: ignore

    def __str__(self):
        return self._output_string

    @property
    def input_string(self) -> str:
        """The input string that initialized the TransductionGraph."""
        return self._input_string

    @input_string.setter
    def input_string(self, value):
        # Only modify this if you're also adjusting the edges at the same time!
        self._input_string = value
        self._input_nodes = list(enumerate(value))

    @property
    def output_string(self) -> str:
        """str: The output string."""
        return self._output_string

    @output_string.setter
    def output_string(self, value):
        self._output_string = value
        self._output_nodes = list(enumerate(value))

    @property
    def input_nodes(self) -> List[Tuple[int, str]]:
        """List of nodes (index and character string) corresponding to the input"""
        return self._input_nodes

    @input_nodes.setter
    def input_nodes(self, value):
        raise ValueError(
            f"Sorry, you tried to change the input nodes to {value} but they cannot be changed."
        )

    @property
    def output_nodes(self) -> List[Tuple[int, str]]:
        """List of nodes (index and character string) corresponding to the output"""
        return self._output_nodes

    @output_nodes.setter
    def output_nodes(self, value):
        raise ValueError(
            f"Sorry, you tried to change the output nodes to {value} but they cannot be changed directly. Change output_string instead."
        )

    @property
    def edges(self) -> List:
        """List[Tuple[int, Optional[int]]] of edges for the transformation.

        Note that currently this is *not* the same between
        TransductionGraph and CompositeTransductionGraph.
        """
        return self._edges

    @edges.setter
    def edges(self, value):
        self._edges = value

    @property
    def debugger(self) -> List[list]:
        """List[List[dict]]: A list of lists of rules applied during the
        transformation. Useful for debugging.

        Note that currently this is *not* the same between TransductionGraph
        and CompositeTransductionGraph.

        """
        return self._debugger

    @debugger.setter
    def debugger(self, value):
        self._debugger = value

    @property
    def tiers(self) -> List["TransductionGraph"]:  # noqa: F821
        """A list of TransductionGraph objects for each tier in the graph
        (there is one tier)."""
        return [self]

    @tiers.setter
    def tiers(self, value):
        raise ValueError(
            f"Sorry, you tried to change the tiers to {value} but they cannot be changed"
        )

    def pretty_edges(self):
        """List[Tuple[str, str]] of edges resolved to the string nodes.

        Note that this currently *not* the same output between
        TransductionGraph and ComposedTransductionGraph.
        """
        edges = self._edges[:]
        edges.sort(key=lambda x: x[0])
        out_edges: List[Tuple[str, Optional[str]]] = []
        for edge in edges:
            assert edge[0] is not None  # Empty inputs are not allowed
            if edge[1] is None:
                out_edges.append((self._input_nodes[edge[0]][1], None))
            else:
                out_edges.append(
                    (
                        self._input_nodes[edge[0]][1],
                        self._output_nodes[edge[1]][1],
                    )
                )
        return out_edges

    def alignments(self) -> List[Tuple[int, Optional[int]]]:
        """Alignments (input node index, output node index) for the full
        (possibly composed) transduction.

        Insertions are not allowed in rules, but deletions are.  In
        most cases they will *not* be represented as `None` in the
        output node index, except in the case where there is a
        deletion and the output string is empty.

        Nonetheless code must be robust to the possibility of `None`
        occurring as the output of an alignment.

        Note that this method *is* compatible between TransductionGraph
        and CompositeTransductionGraph.

        """
        return self._edges

    def substring_alignments(self) -> List[Tuple[str, str]]:  # noqa: C901
        """Alignments of input to output substrings for this graph.

        As opposed to `pretty_edges`, this method will return the
        one-to-many or many-to-one alignments produced by a
        conversion.  It is also the same between TransductionGraph and
        CompositeTransductionGraph.  For example, assuming this input
        and output with these edges::

            ABCDEFF
            aabbcdef
            [(0, 0), (0, 1), (1, 2), (1, 3), (2, 4), (3, 5),
             (4, 6), (5, 7), (6, 7)]

        it should return::

            [('A', 'aa'), ('B', 'bb'), ('C', c'), ('D', 'd'),
             ('E', 'e'), ('FF', 'f')]

        In the case of reorderings, the substrings returned correspond
        to the minimal *monotonic* alignments, that is, the minimal
        subsequences which preserve ordering between the input and
        output.  So, for example given these edges::

            ABCDEF
            abefcd
            [(0, 0), (1, 1), (2, 4), (3, 5), (4, 2), (5, 3)]

        it should return::

            [('A', 'a'), ('B', 'b'), ('CDEF', 'efcd')]

        This means, that in the case of truly radical reorderings,
        there is no "alignment" per se.  This is highly unlikely to
        occur, given the constrained nature of g2p rules.

        Note that in the case of multi-character outputs such as in
        ARPABET, the output substrings *do not necessarily* correspond
        to tokens, and spaces are treated as any other character in
        the alignment (thus, alignments of (' ', ' ') are possible).
        If you wish to reconstruct the input it is your resonsibility
        to collect all the alignments corresponding to an output token
        (this should not be too hard to do).

        """
        alignments = self.alignments()

        def find_monotonic_segments(alignments):
            segments = []
            # Sort according to input and output (None is not allowed on input)
            isort = sorted(
                alignments, key=lambda x: (x[0], x[0]) if x[1] is None else x
            )
            osort = sorted(
                alignments, key=lambda x: (x[0], x[0]) if x[1] is None else (x[1], x[0])
            )
            # print("isort:", isort)
            # print("osort:", osort)
            # Use -1 as flag value because None has a meaning in alignments
            istart = ostart = iend = oend = -1
            for iedge, oedge in zip(isort, osort):
                # Create a new segment if the sorting orders agree, or
                # if the subsequent alignments cannot overlap with the
                # current segment
                non_overlapping = (
                    -1 not in (iend, oend) and iedge[0] > iend and oedge[0] > oend
                )
                if iedge == oedge or non_overlapping:
                    # Output an existing segment if we have one
                    if iend != -1:
                        segments.append((istart, iend, ostart, oend))
                        istart = ostart = iend = oend = -1
                    # And if we have entered an agreement region output that
                    if iedge == oedge:
                        ipos, opos = iedge
                        segments.append((ipos, ipos, opos, opos))
                        istart = ostart = iend = oend = -1
                        continue
                if istart == -1:  # Start a new segment
                    # Smallest input index (by definition)
                    istart = iedge[0]
                    # Smallest output index (by definition)
                    ostart = oedge[1]
                    # These are updated below
                    iend = oedge[0]
                    oend = iedge[1]
                else:  # Expand an existing segment
                    assert oedge[0] is not None
                    assert iedge[1] is not None
                    iend = max(iend, oedge[0])
                    # Replace None with not-None
                    if oend is None:
                        oend = iedge[1]
                    elif iedge[1] is not None:
                        oend = max(oend, iedge[1])
            if istart != -1:
                assert iend != -1
            # Output a final segment if one exists
            if iend != -1:
                segments.append((istart, iend, ostart, oend))
            return segments

        # Construct areas of agreement
        segments = find_monotonic_segments(alignments)
        # print("segments:", segments)

        def merge_overlapping_segments(segments):
            if len(segments) <= 1:
                return segments
            istart, iend, ostart, oend = segments[0]
            output = []
            for seg in segments[1:]:
                # If *start* points are outside current segment, start a new one
                if None in (seg[2], oend):  # handle only None in outputs
                    output_outside = not (None, None) == (seg[2], oend)
                else:
                    output_outside = seg[2] > oend
                if seg[0] > iend and output_outside:
                    output.append((istart, iend, ostart, oend))
                    istart, iend, ostart, oend = seg
                else:
                    # Update endpoints to include this segment
                    iend = seg[1]
                    oend = seg[3]
            output.append((istart, iend, ostart, oend))
            return output

        # Merge overlapping segments
        segments = merge_overlapping_segments(segments)
        # print("merged:", segments)
        # Output string alignments
        substrings = []
        for istart, iend, ostart, oend in segments:
            istr = self._input_string[istart : iend + 1]
            if ostart is None:
                assert oend is None
                ostr = ""
            else:
                ostr = self._output_string[ostart : oend + 1]
            substrings.append((istr, ostr))
        return substrings

    def as_dict(self) -> dict:
        return {
            "edges": self._edges,
            "input": self._input_string,
            "output": self._output_string,
            "input_nodes": self._input_nodes,
            "output_nodes": self._output_nodes,
        }

    def clear_debugger(self):
        self._debugger = []

    def append(self, tg):
        """Append the nodes, edges, strings and debugger from tg to self,
        shifting indices so tg nodes and edges are added after those of self.
        """
        in_offset = len(self._input_nodes)
        out_offset = len(self._output_nodes)
        # append input and output strings
        self._input_string += tg._input_string
        self._output_string += tg._output_string
        # append nodes
        self._input_nodes += [(i + in_offset, x) for (i, x) in tg._input_nodes]
        self._output_nodes += [(i + out_offset, x) for (i, x) in tg._output_nodes]
        # append edges and normalize
        self._edges = normalize_edges(
            self._edges
            + [
                (i + in_offset, None if j is None else j + out_offset)
                for i, j in tg.edges
            ]
        )
        # append debuggers
        self._debugger += tg._debugger

    def __iadd__(self, tg):
        self.append(tg)
        return self


class Transducer(BaseTransducer):
    """This is the fundamental class for performing conversions in the g2p library.

    Each Transducer must be initialized with a Mapping object. The Transducer
    object can then be called to apply the rules from Mapping on a given input.

    Attributes:
        mapping (Mapping): Formatted input/output pairs using the g2p.mappings.Mapping class.
    """

    def __init__(self, mapping: Mapping):
        self.mapping = mapping
        self.case_sensitive = mapping.case_sensitive
        self.preserve_case = mapping.preserve_case
        self.norm_form = mapping.norm_form
        self.out_delimiter = mapping.out_delimiter
        self._index_match_pattern = re.compile(r"(?<={)\d+(?=})")
        self._char_match_pattern = re.compile(r"[^0-9\{\}]+(?={\d+})", re.U)
        self.neural_handlers = {
            str(NEURAL_MODEL_HANDLER.deepphonemizer): deep_phonemizer_handler
        }

    def __repr__(self):
        return f"{self.__class__} between {self.mapping.in_lang} and {self.mapping.out_lang}"

    def __call__(self, to_convert: str):
        """The basic method to transduce an input. A proxy for self.apply_rules.

        Args:
            to_convert (str): The string to convert.

        Returns:
            TransductionGraph: Returns an object with all the nodes representing input
            and output characters and their corresponding edges representing the indices
            of the transformation.
        """
        tg = self.apply_rules(to_convert)
        if self.preserve_case:
            return preserve_case(tg, self.mapping.case_equivalencies)
        else:
            return tg

    @staticmethod
    def _pua_to_index(string: str) -> int:
        """Given a string using with characters in the Supllementary Private Use Area A Unicode block
           Produce the number corresponding to the offset from the beginning of the block.

        Args:
            string (str): The string to convert

        Returns:
            int: The offset from the beginning of the block.
        """
        if string:
            intermediate_ord = ord(string[0])
            return intermediate_ord - 983040
        else:
            return -1

    @property
    def in_lang(self) -> str:
        """Input language node name"""
        return self.mapping.in_lang

    @property
    def out_lang(self) -> str:
        """Output language node name"""
        return self.mapping.out_lang

    @property
    def transducers(self) -> List["Transducer"]:  # noqa: F821
        """Sequence of underlying Transducer objects"""
        return [self]

    def resolve_intermediate_chars(self, output_string) -> str:
        """Go through all chars and resolve any intermediate characters from the Private Supplementary Use Area
        to their mapped equivalents.
        """

        indices_seen: Dict[int, int] = defaultdict(int)
        for i, char in enumerate(output_string):
            intermediate_index = self._pua_to_index(char)
            # if not Private Supplementary Use character
            if intermediate_index < 0:
                continue
            output_char_index = indices_seen[intermediate_index]
            try:
                output_string = (
                    output_string[:i]
                    + strip_index_notation(
                        self.mapping.rules[intermediate_index].rule_output
                    )[output_char_index]
                    + output_string[i + 1 :]
                )
            except IndexError:
                indices_seen[intermediate_index] = 0
                output_char_index = 0
                output_string = (
                    output_string[:i]
                    + strip_index_notation(
                        self.mapping.rules[intermediate_index].rule_output
                    )[output_char_index]
                    + output_string[i + 1 :]
                )
            indices_seen[intermediate_index] += 1
        return output_string

    def get_match_groups(
        self,
        tg: TransductionGraph,
        start_end: Tuple[int, int],
        io: Rule,
        diff_from_input: DefaultDict,
        out_string: str,
        output_start: int,
    ) -> Tuple[dict, dict]:
        """Take the inputs to explicit indices matching and create groups of
            Input and Output matches that are grouped by their explicit indices.

            For example, applying a rule that is defined: a{1}b{2} → b{2}a{1} on the input "ab"
            will return inputs, outputs where:

            inputs = {'1': [{'index': 0, 'string': 'a'}], '2': [{'index': 1, 'string': 'b'}] }
            outputs = {'1': [{'index': 0, 'string': 'b'}], '2': [{'index': 1, 'string': 'a'}] }

            This allows input match groups to be iterated through in sequence regardless of their
            character sequence.

        Args:
            tg (TransductionGraph): the graph holding information about the transduction
            start_end (Tuple(int, int)): a tuple containing the start and end of the input match
            io (Rule): an input/output rule
            diff_from_input (DefaultDict): A dictionary containing the single character distance
                                           from a given character index to its input
            out_string (str): the raw output string
            output_start (int): the diff-offset start of the match with respect to the output

        Returns:
            inputs (dict): dictionary containing matches grouped by explicit index match
            outputs (dict): dictionary containing matches grouped by explicit index match
        """
        input_char_matches = [
            x.group() for x in self._char_match_pattern.finditer(io.rule_input)
        ]
        input_match_indices = [
            x.group() for x in self._index_match_pattern.finditer(io.rule_input)
        ]
        inputs: Dict[str, List[dict]] = {}
        index = 0

        input_start = (
            start_end[0] - diff_from_input[self.get_input_from_output(tg, start_end[0])]
        )

        for i, m in enumerate(input_match_indices):
            for char in input_char_matches[i]:
                if m in inputs:
                    inputs[m].append({"index": index + input_start, "string": char})
                else:
                    inputs[m] = [{"index": index + input_start, "string": char}]
                index += 1
        output_char_matches = [
            x.group() for x in self._char_match_pattern.finditer(out_string)
        ]
        output_match_indices = [
            x.group() for x in self._index_match_pattern.finditer(out_string)
        ]
        outputs: Dict[str, List[dict]] = {}
        index = 0
        for i, m in enumerate(output_match_indices):
            for char in output_char_matches[i]:
                if m in outputs:
                    outputs[m].append({"index": index + output_start, "string": char})
                else:
                    outputs[m] = [{"index": index + output_start, "string": char}]
                index += 1
        return inputs, outputs

    def delete_character(self, tg, index_to_delete, ahh):
        """Delete character at `index_to_delete` in TransductionGraph output

        Args:
            tg (TransductionGraph): the current Transduction Graph
            index_to_delete (int): index of character to delete
            ahh (int): current value of i in calling loop
        """
        # delete character
        tg.output_string = (
            tg.output_string[:index_to_delete] + tg.output_string[index_to_delete + 1 :]
        )
        # update indices
        for k, edge in enumerate(tg.edges):
            if (
                edge[1] is not None
                and (ahh == 0 or tg.edges[k - 1][1] is None)
                and edge[1] == index_to_delete
            ):
                tg.edges[k] = (edge[0], None)
            elif edge[1] is not None and edge[1] >= index_to_delete:
                tg.edges[k] = (edge[0], edge[1] - 1)

    def insert_character(self, tg, character_to_insert, index_to_insert_character):
        """Insert character at `index_to_insert_character` in TransductionGraph output

        Args:
            tg (TransductionGraph): the current Transduction Graph
            character_to_insert (str): the character to insert
            index_to_insert_character (int): index of character to insert
        """
        assert len(character_to_insert) == 1
        tg.output_string = (
            tg.output_string[:index_to_insert_character]
            + character_to_insert
            + tg.output_string[index_to_insert_character:]
        )
        for j, edge in enumerate(tg.edges):
            if edge[1] is not None and edge[1] >= index_to_insert_character:
                tg.edges[j] = (edge[0], edge[1] + 1)

    def change_character(self, tg, character, index_to_change):
        """Change character at `index_to_change` in TransductionGraph output to `character`

        Args:
            tg (TransductionGraph): the current Transduction Graph
            character (str): the character to change to
            index_to_change (int): index of character to change
        """
        assert len(character) == 1
        tg.output_string = (
            tg.output_string[:index_to_change]
            + character
            + tg.output_string[index_to_change + 1 :]
        )

    def update_explicit_indices(
        self, tg, match, start_end, io, diff_from_input, diff_from_output, out_string
    ):
        r"""Takes an arbitrary number of input & output strings and their corresponding index offsets.
        It then zips them up according to the provided indexing notation.

        Example:
            A rule that turns a sequence of k\u0313 to 'k might would have a default indexing of k -> ' and \u0313 -> k
            It might be desired though to show that k -> k and \u0313 -> ' and their indices were transposed.
            For this, the Mapping could be given the following: [{'in': 'k{1}\u0313{2}', 'out': "'{2}k{1}"}]
            Indices are found with r'(?<={)\d+(?=})' and characters are found with r'[^0-9\{\}]+(?={\d+})'
        """
        output_start = start_end[0] - diff_from_output[start_end[0]]
        inputs, outputs = self.get_match_groups(
            tg, start_end, io, diff_from_input, out_string, output_start
        )
        out_string = strip_index_notation(out_string)
        # keep track of deletions that haven't yet been processed by diff_from_x
        deleted = 0
        for match_index, input_matches in inputs.items():
            try:
                output_matches = outputs[match_index]
            except KeyError:
                output_matches = []
            process, longest, shortest = self.get_longest_and_shortest(
                input_matches, output_matches
            )
            for i, char in enumerate(longest):
                # deleted segments don't have indices
                if process != "delete" or i <= len(shortest) - 1:
                    output_index = (
                        output_matches[i]["index"]
                        + diff_from_output[output_matches[i]["index"]]
                    )
                # do basic transduction
                if i <= len(shortest) - 1:
                    if output_index >= len(match.group()) + output_start:
                        output_index = len(match.group()) + output_start - 1
                    self.change_character(tg, output_matches[i]["string"], output_index)
                    # this is needed for metathesis,
                    # but metathesis is only expressable by using explicit indices,
                    # so we fix these indices outside of the change_character method
                    tg.edges = [x for x in tg.edges if x[1] != output_index]
                    tg.edges.append([input_matches[i]["index"], output_index])
                elif process == "insert":
                    self.insert_character(tg, char["string"], output_index)
                    # then add insertion edge
                    tg.edges.append([input_matches[-1]["index"], output_index])
                elif process == "delete":
                    if output_matches:
                        index_to_delete = output_matches[-1]["index"] + i - deleted
                    # if there is no output_matches
                    else:
                        index_to_delete = (
                            input_matches[i]["index"]
                            + diff_from_output[input_matches[i]["index"]]
                            - deleted
                        )
                    self.delete_character(tg, index_to_delete, i)
                    deleted += 1
                    if output_matches:
                        tg.edges = [x for x in tg.edges if x[1] != index_to_delete]
                        tg.edges.append([input_matches[i]["index"], None])

    def get_input_from_output(self, tg, output_node):
        return max(x[0] for x in tg.edges if x[1] == output_node)

    def get_longest_and_shortest(
        self, in_string_or_matches, out_string_or_matches
    ) -> Tuple[str, Union[str, list], Union[str, list]]:
        """Given two strings or match lists determine the longest and shortest. If
           the input is longer than the output, the process is to delete,
           if the output is longer than the input, the process is to insert.
           If the input and output are the same length, the process is basic.

        Args:
            in_string_or_matches (str|List): input string
            out_string_or_matches (str|List): output string
        """
        in_length = len(in_string_or_matches)
        out_length = len(out_string_or_matches)
        if in_length > out_length:
            return "delete", in_string_or_matches, out_string_or_matches
        elif in_length < out_length:
            return "insert", out_string_or_matches, in_string_or_matches
        else:
            return "basic", out_string_or_matches, in_string_or_matches

    def update_default_indices(
        self,
        tg,
        match_start,
        diff_from_output,
        in_string,
        out_string,
    ):
        process, longest, shortest = self.get_longest_and_shortest(
            in_string, out_string
        )

        deleted = 0
        for i, char in enumerate(longest):
            output_index = i + match_start + diff_from_output[i + match_start]
            # if the shorter string still has that output:
            #   keep that index, and convert the character
            if i <= len(shortest) - 1:
                self.change_character(tg, out_string[i], output_index)
            # if the output string is longer than the input string
            # then it is an insertion and we should:
            #  - increment every edge after the insertion
            #  - connect every input edge connected to the previous output to that new insertion
            elif process == "insert":
                self.insert_character(tg, char, output_index)
                # add insertion edge
                for edge in tg.edges:
                    if edge[1] == output_index - 1:
                        tg.edges.append([edge[0], output_index])
            # if the input string is longer than the output string
            # then it is a deletion and we should:
            #  - turn the edge to the deleted index to None if there are no non-None preceding characters
            #  - decrement edges following the deleted character
            elif process == "delete":
                # Nodes
                index_to_delete = output_index - deleted
                self.delete_character(tg, index_to_delete, i)
                deleted += 1

    def apply_unidecode(self, to_convert: str) -> TransductionGraph:
        to_convert = unicode_escape(to_convert)
        saved_to_convert = to_convert
        if self.norm_form:
            to_convert, norm_indices = normalize_with_indices(
                to_convert, self.norm_form
            )
        else:
            norm_indices = None
        tg = TransductionGraph(to_convert)

        # Conversion is done character by character using unidecode
        # We retain spaces in the input, but spaces from unidecode are removed
        converted = []
        for rule_input in to_convert:
            unidecode_str = text_unidecode.unidecode(
                unicodedata.normalize("NFKC", rule_input)
            )
            cc = [
                c
                for c in unidecode_str
                if c.isalpha() or c in UNIDECODE_SPECIALS or rule_input.isspace()
            ]
            converted.append("".join(cc))
        tg.output_string = "".join(converted)

        # Edges are calculated to follow the conversion step by step
        if not tg.output_string:
            # Some inputs get completely deleted by unidecode, in which case there are no
            # valid edges to output.
            tg.edges = []
        else:
            edges = []
            x_len, y_len = 0, 0
            for tgt in converted:
                if tgt:
                    for _ in tgt:
                        edges.append((x_len, y_len))
                        y_len += 1
                else:
                    edges.append((x_len, max(y_len - 1, 0)))
                x_len += 1
            if norm_indices:
                tg.edges = compose_indices(norm_indices, edges)
                tg.input_string = saved_to_convert
            else:
                tg.edges = edges

        return tg

    def apply_lexicon(self, to_convert: str):
        tg = TransductionGraph(to_convert)
        if not self.case_sensitive:
            to_convert = to_convert.lower()
        alignment = find_alignment(self.mapping.alignments, to_convert)
        if not alignment:
            tg.edges = []
            tg.output_string = ""
        else:
            tg.output_string = ""
            edges: List[Tuple[int, int]] = []
            in_pos = 0
            out_pos = 0
            for n_inputs, outtxt in alignment:
                for i in range(n_inputs):
                    for j in range(len(outtxt)):
                        edges.append((in_pos + i, out_pos + j))
                    if len(outtxt) == 0:
                        # Attach deletions to the previous output if
                        # possible, otherwise the next output.  This
                        # will be modified to None below if the output
                        # string is empty.
                        edges.append((in_pos + i, max(0, out_pos - 1)))
                if n_inputs == 0:
                    # Attach insertions to the previous input
                    for j in range(len(outtxt)):
                        edges.append((in_pos, out_pos + j))
                in_pos += n_inputs
                if len(outtxt) != 0:
                    out_pos += len(outtxt) + len(self.out_delimiter)
                    # Be bug-compatible with mappings and add an extra delimiter
                    tg.output_string += outtxt + self.out_delimiter
            # Fix up bogus indices here
            out_len = len(tg.output_string)
            tg.edges = []
            for in_pos, out_pos in edges:
                assert in_pos is not None
                if out_pos >= out_len:
                    tg.edges.append((in_pos, None if out_len == 0 else out_len - 1))
                else:
                    tg.edges.append((in_pos, out_pos))
        return tg

    def apply_rules(self, to_convert: str):  # noqa: C901
        if self.mapping.type == MAPPING_TYPE.unidecode:
            return self.apply_unidecode(to_convert)
        elif self.mapping.type == MAPPING_TYPE.lexicon:
            return self.apply_lexicon(to_convert)
        # perform any normalization TODO-for-review: Are we sure we don't want to apply normalization to the string before the lexicon mapping?
        to_convert = unicode_escape(to_convert)
        saved_to_convert = to_convert
        if not self.case_sensitive:
            to_convert = to_convert.lower()
        if self.norm_form:
            to_convert, norm_indices = normalize_with_indices(
                to_convert, self.norm_form
            )
        else:
            norm_indices = None
        tg = TransductionGraph(to_convert)
        tg.debugger.append([])

        # apply neural transformation after normalization TODO-for-review: maybe the lexicon and unidecode should go here too.
        if self.mapping.type == MAPPING_TYPE.neural:
            neural_handler_id = "-".join(
                [
                    str(self.mapping.neural_model_handler),
                    self.mapping.in_lang,
                    self.mapping.out_lang,
                ]
            )
            # check that the specific model handler is available, otherwise construct it.
            if neural_handler_id not in self.neural_handlers:
                ckpt = download_huggingface_model(
                    self.mapping.hf_model_identifier,
                    filename=self.mapping.hf_model_filepath,
                )
                self.neural_handlers[neural_handler_id] = partial(
                    self.neural_handlers[str(self.mapping.neural_model_handler)],
                    ckpt,
                    self.mapping.neural_kwargs,
                )
            self.update_default_indices(
                tg,
                0,
                defaultdict(int, {n: 0 for n in range(len(tg.output_string))}),
                to_convert,
                self.neural_handlers[neural_handler_id](to_convert),
            )
            return tg
        # initialize values
        intermediate_forms = False
        # iterate rules
        # these variables tracks changes in the output string across processing
        # matches of the same pattern
        diff_from_input = defaultdict(int, {n: 0 for n in range(len(tg.output_string))})
        for io in self.mapping.rules:
            assert isinstance(io, Rule)
            # Do not allow empty rules
            if not io.rule_input and not io.rule_output:
                continue
            io = copy.deepcopy(io)
            # create empty out_string
            out_string = ""
            diff_from_output = defaultdict(
                int, {n: 0 for n in range(len(tg.output_string))}
            )

            matches = reversed(list(io.match_pattern.finditer(tg.output_string)))  # type: ignore
            for match_i, match in enumerate(matches):
                debug_string = tg.output_string
                start = match.start()
                end = match.end()
                if match_i:
                    start += diff_from_output[start]
                    end += diff_from_output[end - 1]
                if io.intermediate_form:
                    out_string = io.intermediate_form
                    intermediate_forms = True
                else:
                    out_string = io.rule_output
                if self.out_delimiter:
                    out_string += self.out_delimiter
                if any(self._char_match_pattern.finditer(io.rule_input)) and any(
                    self._char_match_pattern.finditer(out_string)
                ):
                    self.update_explicit_indices(
                        tg,
                        match,
                        (start, end),
                        io,
                        diff_from_input,
                        diff_from_output,
                        out_string,
                    )
                else:
                    self.update_default_indices(
                        tg,
                        match.start(),
                        diff_from_output,
                        match.group(),
                        out_string,
                    )
                if (
                    io.rule_input != io.rule_output
                    or io.context_after
                    or io.context_before
                ):
                    tg.debugger[-1].append(
                        {
                            "input": debug_string,
                            "output": tg.output_string,
                            "rule": io.export_to_dict(),
                            "start": match.start(),
                            "end": match.end(),
                        }
                    )
                out_string = strip_index_notation(out_string)
                # update the output intermediate diff after each match
                diff = len(out_string) - len(match.group())

                try:
                    input_index = self.get_input_from_output(tg, match.end() - 1)
                except ValueError:
                    # it's been deleted
                    input_index = match.end() - 1
                for n in range(
                    match.end() - 1 + diff,
                    max(len(diff_from_output) + diff, len(diff_from_output)),
                ):
                    diff_from_output[n] += diff
                for n in range(input_index, len(diff_from_input)):
                    diff_from_input[n] += diff

        if intermediate_forms:
            tg.output_string = self.resolve_intermediate_chars(tg.output_string)

        # sort and fix None
        tg.edges = normalize_edges(tg.edges)
        if norm_indices is not None:
            tg.edges = compose_indices(norm_indices, tg.edges)
            tg.input_string = saved_to_convert
        return tg

    def check(  # type: ignore[override]
        self,
        tg: TransductionGraph,
        shallow=False,
        display_warnings=False,
        original_input=None,
    ) -> bool:
        out_lang = self.mapping.out_lang
        if "eng-arpabet" in out_lang:
            if is_arpabet(tg.output_string):
                return True
            if display_warnings:
                display_input = original_input or tg.input_string
                LOGGER.warning(
                    f'Transducer output "{tg.output_string}" for input "{display_input}" '
                    "is not fully valid eng-arpabet as recognized by soundswallower."
                )
            return False
        elif is_ipa(out_lang):
            if is_panphon(tg.output_string, display_warnings=display_warnings):
                return True
            if display_warnings:
                display_input = original_input or tg.input_string
                LOGGER.warning(
                    f'Transducer output "{tg.output_string}" for input "{display_input}" is not fully valid {out_lang}.'
                )
            return False
        else:
            # No check implemented at this tier, just return True
            return True


class CompositeTransductionGraph(TransductionGraph):
    """This is the object returned after performing a transduction using a CompositeTransducer.

    Each CompositeTransductionGraph must be initialized with a list of TransductionGraph objects.
    """

    def __init__(self, tg_list):
        self._initialize(tg_list)

    def _initialize(self, tg_list):
        # Separated form __init
        # Plain strings
        self._input_string = tg_list[0].input_string
        self._output_string = tg_list[-1].output_string
        # Nodes
        self._input_nodes = tg_list[0].input_nodes
        self._output_nodes = tg_list[-1].output_nodes
        # Edges
        self._edges = [x.edges for x in tg_list]
        # Debugger
        self._debugger = [x.debugger for x in tg_list]
        # Tiers
        self._tiers = tg_list

    @property
    def tiers(self) -> List[TransductionGraph]:
        """A list of TransductionGraph objects for each tier in the CompositeTransducer"""
        return self._tiers

    @tiers.setter
    def tiers(self, value):
        raise ValueError(
            f"Sorry, you tried to change the tiers to {value} but they cannot be changed"
        )

    @property
    def edges(self) -> List:
        """List[List[Tuple[int, Optional[int]]]] of edges for each tier in the
        transformation.

        Note that this is unfortunately *not* the same between
        TransductionGraph and ComposedTransductionGraph.
        """
        return self._edges

    @edges.setter
    def edges(self, value):
        raise ValueError(
            f"Sorry, you tried to change the edges to {value} but they cannot be changed"
        )

    @property
    def debugger(self) -> List[list]:
        """List[List[List[dict]]]: A list of lists of lists of rules applied
        during the transformation. Useful for debugging.  Note that
        this is not the same between TransductionGraph and
        CompositeTransductionGraph.

        """
        return self._debugger

    @debugger.setter
    def debugger(self, value):
        raise ValueError(
            f"Sorry, you tried to change the debugger to {value} but they cannot be changed"
        )

    def pretty_edges(self):
        """List[List[Tuple[str, str]]] of edges for each tier, expressed as
        strings.

        Note that this is unfortunately *not* the same output between
        TransductionGraph and ComposedTransductionGraph.

        """
        pretty_edges = []
        for tier in self._tiers:
            pretty_edges.append(tier.pretty_edges())
        return pretty_edges

    def alignments(self) -> List[Tuple[int, Optional[int]]]:
        """Return list of alignments (input node index, output node index) for
        the full transduction.

        Note that this *is* the same between TransductionGraph and
        CompositeTransductionGraph.
        """
        composed = self.tiers[0].edges
        for tier in self.tiers[1:]:
            composed = compose_indices(composed, tier.edges)
        return composed

    def as_dict(self):
        return {
            "edges": self._edges,
            "input": self._input_string,
            "output": self._output_string,
            "input_nodes": self._input_nodes,
            "output_nodes": self._output_nodes,
        }

    def append(self, tg):
        if isinstance(tg, CompositeTransductionGraph):
            assert len(self._tiers) == len(tg._tiers)
            for i in range(len(self._tiers)):
                self._tiers[i].append(copy.deepcopy(tg.tiers[i]))
        else:
            for tier in self._tiers:
                tier.append(copy.deepcopy(tg))
        self._initialize(self.tiers)

    def __iadd__(self, tg):
        self.append(tg)
        return self

    def clear_debugger(self):
        self._debugger = []
        for tier in self._tiers:
            tier.clear_debugger()


class CompositeTransducer(BaseTransducer):
    """This class combines Transducer objects to form a CompositeTransducer object.

    Attributes:
        transducers (List[Transducer]): A list of Transducer objects to compose.
    """

    def __init__(self, transducers: List[Transducer]):
        self._transducers = transducers
        self.norm_form = transducers[0].norm_form if transducers else "none"

    def __repr__(self):
        return f"{self.__class__} between {self._transducers[0].mapping.in_lang} and {self._transducers[-1].mapping.out_lang}"

    def __call__(self, to_convert: str):
        return self.apply_rules(to_convert)

    @property
    def transducers(self):
        """Sequence of underlying Transducer objects"""
        return self._transducers

    @property
    def in_lang(self):
        """Input language node name"""
        return self._transducers[0].in_lang

    @property
    def out_lang(self):
        """Output language node name"""
        return self._transducers[-1].out_lang

    def apply_rules(self, to_convert: str):
        tg_list = []
        for transducer in self._transducers:
            tg = transducer(to_convert)
            tg_list.append(tg)
            to_convert = tg.output_string
        return CompositeTransductionGraph(tg_list)

    def check(  # type: ignore[override]
        self, tg: CompositeTransductionGraph, shallow=False, display_warnings=False
    ) -> bool:
        assert len(self._transducers) == len(tg._tiers)
        if shallow:
            return self._transducers[-1].check(
                tg._tiers[-1],
                display_warnings=display_warnings,
                original_input=tg.input_string,
            )
        else:
            result = True
            for i, transducer in enumerate(self._transducers):
                if not transducer.check(
                    tg._tiers[i],
                    display_warnings=display_warnings,
                    original_input=tg.input_string,
                ):
                    # Don't short circuit if warnings are required
                    if display_warnings:
                        result = False
                    else:
                        return False
            return result


class TokenizingTransducer(BaseTransducer):
    """This class combines tokenization and transduction.

    This is particularly useful for lexicon mappings, which cannot
    handle more than a single word at a time.

    Attributes:
        transducer (Transducer): A Transducer object for the mapping part
        tokenizer (Tokenizer): A Tokenizer object to split the string before mapping
    """

    def __init__(
        self,
        transducer: Union[Transducer, CompositeTransducer],
        tokenizer: BaseTokenizer,
    ):
        self._transducer = transducer
        self._tokenizer = tokenizer

    def __call__(self, to_convert: str):
        # perform normalization before tokenizing, since it can change tokenization
        if self._transducer.norm_form:
            to_convert = normalize(to_convert, self._transducer.norm_form)

        # Initialize the transducer on an empty string so we can handle inputs
        # that start with a non-token correctly.
        tg = self._transducer("")
        tg.clear_debugger()  # clear the meaningless initial debugger

        for token in self._tokenizer.tokenize_text(to_convert):
            if token.is_word:
                word_tg = self._transducer(token.text)
                tg += word_tg
            else:
                non_word_tg = TransductionGraph(token.text)
                tg += non_word_tg
        return tg

    @property
    def transducer(self):
        """Underlying Transducer object"""
        return self._transducer

    @property
    def transducers(self) -> List[Transducer]:
        """Sequence of underlying Transducer objects"""
        return self._transducer.transducers

    @property
    def in_lang(self) -> str:
        """Input language node name"""
        return self._transducer.in_lang

    @property
    def out_lang(self) -> str:
        """Output language node name"""
        return self._transducer.out_lang

    def check(  # type: ignore[override]
        self, tg: TransductionGraph, shallow=False, display_warnings=False
    ) -> bool:
        # The obvious implementation fails, because we need to check only the words, not
        # the text between the words!
        # return self._transducer.check(tg) # <- complains about characters between words

        # So, sadly, we redo the work of transduction so we can check the words only, step
        # by step. I don't like this solution, but I don't see how to get around it.
        result = True
        for token in self._tokenizer.tokenize_text(tg.input_string):
            if token.is_word and not self._transducer.check(
                self._transducer(token.text),
                shallow,
                display_warnings=display_warnings,
            ):
                # Don't short circuit if warnings are required
                if display_warnings:
                    result = False
                else:
                    return False
        return result


def preserve_case(
    tg: TransductionGraph, case_equivalencies: Optional[Dict[str, str]] = None
) -> TransductionGraph:
    if case_equivalencies is None:
        case_equivalencies = {}
    reverse_case_equivalencies = {v: k for k, v in case_equivalencies.items()}
    all_lower_case_equivalencies = case_equivalencies.keys()
    all_upper_case_equivalencies = case_equivalencies.values()
    new_string = ""
    for item in tg.substring_alignments():
        in_sub = item[0]
        out_sub = item[1]
        any_in_upper = any(x.isupper() for x in in_sub)
        any_in_lower = any(x.islower() for x in in_sub)
        any_out_upper = any(x.isupper() for x in out_sub)
        any_out_lower = any(x.islower() for x in out_sub)
        # continue if character is un-caseable
        if (
            out_sub not in case_equivalencies
            and not any_out_upper
            and not any_out_lower
        ):
            new_string += out_sub
            continue
        # lower case using case equivalencies if they exist
        if (
            any_in_lower or in_sub in all_lower_case_equivalencies
        ) and out_sub in all_upper_case_equivalencies:
            new_string += reverse_case_equivalencies[out_sub]
            continue
        # upper case using case equivalencies if they exist
        elif (
            any_in_upper or in_sub in all_upper_case_equivalencies
        ) and out_sub in all_lower_case_equivalencies:
            new_string += case_equivalencies[out_sub]
            continue
        # change to upper if required
        if any_in_upper and any_out_lower:
            new_string += out_sub.upper()
            continue
        # change to lower if required
        if any_in_lower and any_out_upper:
            new_string += out_sub.lower()
            continue
        # just in case, append the out_sub
        new_string += out_sub
    tg.output_string = new_string
    return tg



"""
    Error Log

"""

import logging
import os

import coloredlogs  # type: ignore

FIELD_STYLES = dict(
    levelname=dict(color="green"),
)


def setup_logger(name):
    """Create logger and configure with cool colors!"""

    logger = logging.getLogger(name)
    coloredlogs.install(
        level=os.environ.get("G2P_LOGLEVEL", "INFO").upper(),
        fmt="%(levelname)s - %(message)s",
        logger=logger,
        field_styles=FIELD_STYLES,
    )
    return logger


LOGGER = setup_logger("root")



#!/usr/bin/env python

from unittest import TestCase, main

import g2p.mappings.tokenizer as tok
from g2p.log import LOGGER


class TokenizerTest(TestCase):
    """Test suite for tokenizing text in a language-specific way"""

    def test_tokenize_fra(self):
        input = "ceci était 'un' test."
        tokenizer = tok.make_tokenizer("fra")
        tokens = tokenizer.tokenize_text(input)
        self.assertEqual(len(tokens), 8)
        self.assertTrue(tokens[0].is_word)
        self.assertEqual(tokens[0].text, "ceci")
        self.assertFalse(tokens[1].is_word)
        self.assertEqual(tokens[1].text, " ")
        self.assertTrue(tokens[2].is_word)
        self.assertEqual(tokens[2].text, "était")
        self.assertFalse(tokens[3].is_word)
        self.assertEqual(tokens[3].text, " '")
        self.assertTrue(tokens[4].is_word)
        self.assertEqual(tokens[4].text, "un")
        self.assertFalse(tokens[5].is_word)
        self.assertEqual(tokens[5].text, "' ")
        self.assertTrue(tokens[6].is_word)
        self.assertEqual(tokens[6].text, "test")
        self.assertFalse(tokens[7].is_word)
        self.assertEqual(tokens[7].text, ".")

    def test_tokenize_eng(self):
        input = "This is éçà test."
        tokenizer = tok.make_tokenizer("eng")
        tokens = tokenizer.tokenize_text(input)
        self.assertEqual(len(tokens), 8)
        self.assertTrue(tokens[0].is_word)
        self.assertEqual(tokens[0].text, "This")
        self.assertFalse(tokens[1].is_word)
        self.assertEqual(tokens[1].text, " ")

    def test_lexicon_tokenizer(self):
        tokenizer = tok.make_tokenizer("eng")
        tests = [
            ("It's", ["It's"]),
            ("'cause", ["'cause"]),
            ('"\'cause"', ['"', "'cause", '"']),
            ("aardvark's", ["aardvark", "'s"]),
            ("'aardvark's'", ["'", "aardvark", "'s", "'"]),
            ("ten a.m.", ["ten", " ", "a.m."]),
            ('ten "a.m.,!"', ["ten", ' "', "a.m.", ',!"']),
            ("all-out war", ["all-out", " ", "war"]),  # all-out is in the lexicon
            ("all-in: nonsense", ["all", "-", "in", ": ", "nonsense"]),  # all-in is not
        ]
        for input_text, expected_tokens in tests:
            with self.subTest(input_text=input_text):
                tokens = tokenizer.tokenize_text(input_text)
                self.assertEqual([x.text for x in tokens], expected_tokens)

    def test_tokenize_win(self):
        """win is easy to tokenize because win -> win-ipa exists and has ' in its inventory"""
        input = "p'ōį̄ą"
        self.assertEqual(len(tok.make_tokenizer("fra").tokenize_text(input)), 3)

        tokenizer = tok.make_tokenizer("win")
        tokens = tokenizer.tokenize_text(input)
        self.assertEqual(len(tokens), 1)
        self.assertTrue(tokens[0].is_word)
        self.assertEqual(tokens[0].text, "p'ōį̄ą")

    def test_tokenize_tce(self):
        """tce is hard to tokenize correctly because we have tce -> tce-equiv -> tce-ipa, and ' is
        only mapped in the latter.
        Challenges:
         - since tce->tce-ipa is not a direct mapping, we're probably getting a default
           tokenizer
         - we want to merge the input inventory of both tce->tce-equiv and tce-equiv->tce-ipa
           into just one joint inventory for the purpose of tokenization.
        Now works - issue #46 fixed this.
        """
        input = "ts'nj"
        self.assertEqual(len(tok.make_tokenizer("fra").tokenize_text(input)), 3)

        tokenizer = tok.make_tokenizer("tce")
        tokens = tokenizer.tokenize_text(input)
        self.assertEqual(len(tokens), 1)
        self.assertTrue(tokens[0].is_word)
        self.assertEqual(tokens[0].text, "ts'nj")

    def test_tokenize_tce_equiv(self):
        input = "ts'e ts`e ts‘e ts’"
        self.assertEqual(len(tok.make_tokenizer("fra").tokenize_text(input)), 14)
        # tce_tokens = tok.make_tokenizer("tce").tokenize_text(input)
        # LOGGER.warning([x.text for x in tce_tokens])
        self.assertEqual(len(tok.make_tokenizer("tce").tokenize_text(input)), 7)

    def test_tokenizer_identity_tce(self):
        self.assertNotEqual(tok.make_tokenizer("eng"), tok.make_tokenizer("fra"))
        self.assertNotEqual(tok.make_tokenizer("eng"), tok.make_tokenizer("tce"))
        self.assertEqual(tok.make_tokenizer("tce"), tok.make_tokenizer("tce"))
        self.assertNotEqual(tok.make_tokenizer("tce"), tok.make_tokenizer())
        self.assertEqual(tok.make_tokenizer("foo"), tok.make_tokenizer())

    def test_tokenize_kwk(self):
        """kwk is easier than tce: we just need to use kwk-umista -> kwk-ipa, but that's not
        implemented yet.
        Now works - issue #46 fixed this.
        """
        self.assertEqual(
            len(tok.make_tokenizer("kwk-umista").tokenize_text("kwak'wala")), 1
        )

    def test_three_hop_tokenizer(self):
        # This used to test the three hop tokenizer with haa -> haa-ipa via haa-equiv and haa-simp
        # tokenizer = tok.make_tokenizer("haa", tok_path=["haa", "haa-equiv", "haa-simp", "haa-ipa"])
        # But now haa has been redesigned to not use haa-simp, so downgrade the test to two hops
        tokenizer = tok.make_tokenizer("haa", tok_path=["haa", "haa-equiv", "haa-ipa"])
        tokens = tokenizer.tokenize_text("ch'ch")
        self.assertEqual(len(tokens), 1)

    def test_tokenize_not_ipa_explicit(self):
        tokenizer = tok.make_tokenizer("fn-unicode-font", "fn-unicode")
        self.assertNotEqual(tokenizer, tok.make_tokenizer())

    def test_tokenize_not_ipa_implicit(self):
        tokenizer = tok.make_tokenizer("fn-unicode-font")
        self.assertNotEqual(tokenizer, tok.make_tokenizer())

    def test_tokenize_lang_does_not_exist(self):
        self.assertEqual(tok.make_tokenizer("not_a_language"), tok.make_tokenizer())
        with self.assertLogs(LOGGER):
            self.assertEqual(
                tok.make_tokenizer("fra", "not_a_language"), tok.make_tokenizer()
            )

    def test_make_tokenizer_error(self):
        with self.assertRaises(ValueError):
            _ = tok.make_tokenizer("fra", "eng-arpabet", ["fra-ipa", "eng-ipa"])

    def test_deprecated_warning(self):
        with self.assertLogs(LOGGER, level="WARNING") as cm:
            tok._deprecated_warning_printed = False
            self.assertEqual(tok.get_tokenizer("fra"), tok.make_tokenizer("fra"))
        self.assertIn("deprecated", "".join(cm.output))

    def test_gwi_multichar_grapheme_makeg2p(self):
        from g2p import make_g2p

        g2p_engine = make_g2p("gwi", "gwi-ipa")
        _ = g2p_engine("ı̨")  # we're just confirming this does not raise, see #430

    def test_gwi_multichar_grapheme_tok(self):
        tokd = tok.make_tokenizer("gwi").tokenize_text("ı̨")
        self.assertEqual("ı̨", tokd[0].text)


if __name__ == "__main__":
    main()



"""
Language mappings for g2p.
"""

import gzip
import json
import os

from g2p.constants import LANGS_DIR, LANGS_FILE_NAME, NETWORK_FILE_NAME
from g2p.log import LOGGER

from .network_lite import DiGraph, node_link_graph

assert LANGS_DIR == os.path.dirname(__file__)
LANGS_PKL = os.path.join(LANGS_DIR, LANGS_FILE_NAME)
LANGS_NWORK_PATH = os.path.join(LANGS_DIR, NETWORK_FILE_NAME)


def load_langs(path: str = LANGS_PKL):
    try:
        with gzip.open(path, "rt", encoding="utf8") as f:
            return json.load(f)
    except Exception as e:
        LOGGER.warning(f"Failed to read language cache from {path}: {e}")
        return {}


def load_network(path: str = LANGS_NWORK_PATH) -> DiGraph[str]:
    try:
        with gzip.open(path, "rt", encoding="utf8") as f:
            data = json.load(f)
            return node_link_graph(data)
    except Exception as e:
        LOGGER.warning(f"Failed to read language network from {path}: {e}")
        return DiGraph()


def get_available_languages(langs: dict) -> list:
    language_names = set()
    for k, v in langs.items():
        if k in ["generated", "font-encodings"]:
            continue
        for vv in v["mappings"]:
            if "language_name" in vv:
                language_names.add(vv["language_name"])
    return sorted(language_names)


def get_available_mappings(langs: dict) -> list:
    mappings_available = []
    for v in langs.values():
        if "mappings" in v:
            mappings_available.extend(v["mappings"])
        else:
            mappings_available.append(v)
    return mappings_available


LANGS_NETWORK = load_network()
# Making private because it should be imported from g2p.mappings instead
_LANGS = load_langs()
LANGS_AVAILABLE = get_available_languages(_LANGS)
_MAPPINGS_AVAILABLE = get_available_mappings(_LANGS)


def reload_db():
    """Reload the langs and network data, necessary after a g2p update."""

    # We update all structures in place, so that another module having done from
    # g2p.mappings.langs import VAR will see the udpates without any code changes.

    global _LANGS
    _LANGS.clear()
    _LANGS.update(load_langs())

    global LANGS_NETWORK
    LANGS_NETWORK.clear()
    new_langs_network = load_network()
    LANGS_NETWORK.update(new_langs_network.edges, new_langs_network.nodes)

    global LANGS_AVAILABLE
    LANGS_AVAILABLE.clear()
    LANGS_AVAILABLE.extend(get_available_languages(_LANGS))

    global _MAPPINGS_AVAILABLE
    _MAPPINGS_AVAILABLE.clear()
    _MAPPINGS_AVAILABLE.extend(get_available_mappings(_LANGS))



#!/usr/bin/env python

import sys
from unittest import TestCase, main, SkipTest

from g2p import make_g2p
from g2p.log import LOGGER
from g2p.tests.public.data import load_neural_test_data


class NeuralLangTest(TestCase):
    """Basic Test for individual lookup tables.

    Test file is in g2p/tests/public/data/neural.psv.

    Each line in the test file consists of SOURCE,TARGET,INPUT,OUTPUT

    """

    def test_io(self):
        if sys.version_info < (3, 9):
            raise SkipTest("neural g2p requires python >= 3.9")
        langs_to_test = load_neural_test_data()

        # go through each language declared in the test case set up
        # Instead of asserting immediately, we go through all the cases first, so that
        # running test_neural.py prints all the errors at once, to help debugging a given g2p mapping.
        # Then we call assertEqual on the first failed case, to make unittest register the failure.
        error_count = 0
        error_prefix = "test_neural.py: mapping error"
        for test in langs_to_test:
            transducer = make_g2p(test[0], test[1], neural=True)
            output_string = transducer(test[2]).output_string.strip()
            if output_string != test[3].strip():
                LOGGER.error(
                    "{} for {}: {} from {} to {} should be {}, got {}".format(
                        error_prefix,
                        test[-1],
                        test[2],
                        test[0],
                        test[1],
                        test[3],
                        output_string,
                    )
                )
                error_count += 1

        self.assertEqual(
            error_count,
            0,
            f'Search for "ERROR - {error_prefix}" above to find all the g2p mapping errors.',
        )


if __name__ == "__main__":
    main()



import datetime as dt

from text_unidecode import unidecode  # type: ignore

from g2p import make_g2p
from g2p.log import LOGGER
from g2p.mappings import Mapping
from g2p.mappings.create_ipa_mapping import align_inventories
from g2p.mappings.utils import is_ipa, unicode_escape

DUMMY_INVENTORY = ["ɑ", "i", "u", "t", "s", "n"]


def align_to_dummy_fallback(
    mapping: Mapping,
    io: str = "in",
    distance: str = "weighted_feature_edit_distance",
    quiet=False,
):
    """Create a mapping from mapping's output inventory to a minimalist dummy inventory"""
    mapping_config = mapping.model_dump()
    config = {
        "in_lang": mapping_config[f"{io}_lang"],
        "out_lang": "dummy",
        "authors": [f"Generated {dt.datetime.now()}"],
    }
    default_char = "t"
    if is_ipa(mapping_config[f"{io}_lang"]):
        list_of_rules = align_inventories(
            mapping.inventory(io), DUMMY_INVENTORY, distance=distance, quiet=quiet
        )
    else:
        und_g2p = make_g2p("und", "und-ipa", tokenize=False)
        list_of_rules = [
            {
                "in": unicode_escape(x),
                "out": und_g2p(unidecode(x).lower()).output_string,
            }
            for x in mapping.inventory(io)
        ]
        dummy_list = align_inventories(
            [x["out"] for x in list_of_rules],
            DUMMY_INVENTORY,
            distance=distance,
            quiet=quiet,
        )
        dummy_dict = {}
        for x in dummy_list:
            if x["in"]:
                dummy_dict[x["in"]] = x["out"]

        for x in list_of_rules:
            try:
                x["out"] = dummy_dict[x["out"]]
            except KeyError:
                LOGGER.warning(
                    f"We couldn't guess at what {x['in']} means, so it's being "
                    f"replaced with '{default_char}' instead."
                )
                x["out"] = default_char

    config["rules"] = list_of_rules
    return Mapping(**config)



"""

Utilities used by other classes

"""

import gzip
import json
from pathlib import Path
from typing import Any, Optional

from g2p.log import LOGGER
from g2p.mappings import MAPPINGS_AVAILABLE, Mapping, MappingConfig
from g2p.mappings.langs import LANGS_DIR, LANGS_NETWORK, LANGS_NWORK_PATH, LANGS_PKL
from g2p.mappings.utils import MAPPING_TYPE, is_ipa

from .network_lite import DiGraph, node_link_data

# panphon.distance.Distance() takes a long time to initialize, so...
# a) we don't want to load it if we don't need it, i.e., don't use a constant
# b) we don't want to load it more than once, i.e., don't use a local variable.
# Conclusion: use a singleton with lazy initialization
# Profiling results: calling is_panphon() the first time still costs 400ms, but subsequent
# calls cost .02ms each. With dst=panphone.distance.Distance() instead of
# dst=getPanphonDistanceSingleton(), the first call costs 400ms and subsequent calls
# cost 180ms each.

_PANPHON_DISTANCE_SINGLETON = None


def getPanphonDistanceSingleton():
    global _PANPHON_DISTANCE_SINGLETON
    if _PANPHON_DISTANCE_SINGLETON is None:
        # Expensive import, only do it when actually needed
        import panphon.distance  # type: ignore

        _PANPHON_DISTANCE_SINGLETON = panphon.distance.Distance()
    return _PANPHON_DISTANCE_SINGLETON


def check_ipa_known_segs(mappings_to_check=False) -> bool:
    """Check the given mappings, or all IPA mappings, for invalid IPA in the "out" fields

    Returns True iff not errors were found.
    """
    if not mappings_to_check:
        mappings_to_check = [x.out_lang for x in MAPPINGS_AVAILABLE]
    found_error = False

    for mapping in [x for x in MAPPINGS_AVAILABLE if x.out_lang in mappings_to_check]:
        if is_ipa(mapping.out_lang) and mapping.type == MAPPING_TYPE.mapping:
            reverse = mapping.reverse
            for rule in mapping.rules:
                output = rule.rule_input if reverse else rule.rule_output
                if not is_panphon(output):
                    LOGGER.warning(
                        f"Output '{rule.rule_output}' in rule {rule} in mapping between {mapping.in_lang} "
                        f"and {mapping.out_lang} is not recognized as valid IPA by panphon."
                    )
                    found_error = True
    if found_error:
        LOGGER.warning(
            "Please refer to https://github.com/dmort27/panphon for information about panphon."
        )
    return not found_error


_is_panphon_g_warning_printed = False
_is_panphon_colon_warning_printed = False


def is_panphon(string, display_warnings=False):
    # Deferred importing required here, because g2p.transducer also imports this file.
    # Such circular dependency is probably bad design, maybe a reviewer of this code will
    # have a better solution to recommend?
    import g2p.transducer

    dst = getPanphonDistanceSingleton()
    panphon_preprocessor = g2p.transducer.Transducer(
        Mapping.find_mapping_by_id("panphon_preprocessor")
    )
    preprocessed_string = panphon_preprocessor(string).output_string
    # Use a loop that prints the warnings on all strings that are not panphon, even though
    # logically this should not be necessary to calculate the answer.
    result = True
    for word in preprocessed_string.split():
        word_ipa_segs = dst.fm.ipa_segs(word)
        word_ipa = "".join(word_ipa_segs)
        if word != word_ipa:
            if not display_warnings:
                return False
            LOGGER.warning(
                f'String "{word}" is not identical to its IPA segmentation: {word_ipa_segs}'
            )
            global _is_panphon_g_warning_printed
            if "g" in word and not _is_panphon_g_warning_printed:
                LOGGER.warning(
                    "Common IPA gotcha: the ASCII 'g' character is not IPA, use 'ɡ' (\\u0261) instead."
                )
                _is_panphon_g_warning_printed = True
            global _is_panphon_colon_warning_printed
            if ":" in word and not _is_panphon_colon_warning_printed:
                LOGGER.warning(
                    "Common IPA gotcha: the ASCII ':' character is not IPA, use 'ː' (\\u02D0) instead."
                )
                _is_panphon_colon_warning_printed = True
            for c in word:
                if c not in word_ipa:
                    LOGGER.warning(
                        f"Character '{c}' (\\u{format(ord(c), '04x')}) in word '{word}' "
                        "was not recognized as IPA by panphon."
                    )
            result = False
    return result


_ARPABET_SET = None


def is_arpabet(string):
    global _ARPABET_SET
    if _ARPABET_SET is None:
        _ARPABET_SET = set(
            Mapping.find_mapping(in_lang="eng-ipa", out_lang="eng-arpabet").inventory(
                "out"
            )
        )
    # print(f"arpabet_set={_ARPABET_SET}")
    for sound in string.split():
        if sound not in _ARPABET_SET:
            return False
    return True


def cache_langs(
    dir_path: str = LANGS_DIR,
    langs_path: str = LANGS_PKL,
    network_path: str = LANGS_NWORK_PATH,
):
    """Read in all files and save as pickle.

    Args:
       dir_path: Path to scan for config-g2p.yaml files.  Default is the
                 installed g2p/mappings/langs directory.
       langs_path: Path to output langs.json.gz file.  Default is
                   the installed g2p/mappings/langs/langs.json.gz
       network_path: Path to output pickle file.  Default is the
                     installed g2p/mappings/langs/network.pkl.
    """
    langs = {}

    # Sort by language code
    paths = sorted(
        Path(dir_path).glob("./*/config-g2p.y*ml"), key=lambda x: x.parent.stem
    )
    mappings_legal_pairs = []
    if not paths:
        raise FileNotFoundError(
            f"There don't seem to be any valid mappings in {dir_path}"
        )
    for path in paths:
        code = path.parent.stem
        mapping_config = MappingConfig.load_mapping_config_from_path(path)
        # TODO: should put in some measure to prioritize non-generated
        # mappings and warn when they override
        mappings_legal_pairs.extend(
            [(mapping.in_lang, mapping.out_lang) for mapping in mapping_config.mappings]
        )
        langs[code] = mapping_config.export_to_dict()

    # Save as a Directional Graph
    lang_network: DiGraph[str] = DiGraph()
    lang_network.add_edges_from(mappings_legal_pairs)
    write_json_gz(network_path, node_link_data(lang_network))
    write_json_gz(langs_path, langs)
    return langs


def write_json_gz(path: str, data: Any):
    with gzip.GzipFile(path, "wb", mtime=0) as zipfile:
        zipfile.write(
            json.dumps(
                data,
                separators=(",", ":"),
                ensure_ascii=False,
                sort_keys=True,
            ).encode("utf-8")
        )


def network_to_echart(outfile: Optional[str] = None, layout: bool = False):
    nodes = []
    no_nodes = len(LANGS_NETWORK.nodes)
    for node in LANGS_NETWORK.nodes:
        lang_name = node.split("-")[0]
        no_ancestors = len(LANGS_NETWORK.ancestors(node))
        no_descendants = len(LANGS_NETWORK.descendants(node))
        size = min(
            20,
            max(
                2, ((no_ancestors / no_nodes) * 100 + (no_descendants / no_nodes) * 100)
            ),
        )
        size = round(size, 2)
        node = {"name": node, "symbolSize": size, "id": node, "category": lang_name}
        nodes.append(node)
    nodes.sort(key=lambda x: x["name"])
    edges = []
    for edge in LANGS_NETWORK.edges:
        edges.append({"source": edge[0], "target": edge[1]})
    if outfile:
        with open(
            outfile,
            "w",
            encoding="utf-8",
            newline="\n",
        ) as f:
            f.write(json.dumps({"nodes": nodes, "edges": edges}) + "\n")
        LOGGER.info("Wrote network nodes and edges to static file.")
    return nodes, edges



"""
All custom Exceptions
"""


# traceback from exceptions that inherit from this class are suppressed
class CommandLineError(Exception):
    """The traceback of all CommandLineError's is supressed when the
    errors occur on the command line to provide a useful command line
    interface.
    """

    def render(self, msg):
        return msg % vars(self)


class NeuralDependencyError(ImportError):
    def __init__(self, func_name=None):
        if func_name:
            message = (
                f"Function '{func_name}' requires neural dependencies. "
                "Install with: pip install g2p[neural]"
            )
        else:
            message = "This operation requires neural dependencies. Install with: pip install g2p[neural]"
        super().__init__(message)


class MappingMissing(CommandLineError):
    def __init__(self, in_lang, out_lang):
        super().__init__(self)
        self.in_lang = in_lang
        self.out_lang = out_lang

    def __str__(self):
        return self.render(
            (
                "\n"
                'There is no mapping between the languages "%(in_lang)s" and "%(out_lang)s", \n'
                "please make sure you spelled the name correctly or go to\n"
                "https://g2p-studio.herokuapp.com/api/v1/langs for a list of mappings"
            )
        )


class NoPath(CommandLineError):
    def __init__(self, in_lang, out_lang):
        super().__init__(self)
        self.in_lang = in_lang
        self.out_lang = out_lang

    def __str__(self):
        return self.render(
            (
                "\n"
                'There is no g2p path between the languages "%(in_lang)s" and "%(out_lang)s", \n'
                "please make sure you spelled the name correctly or go to\n"
                "https://g2p-studio.herokuapp.com/api/v1/langs for a list of mappings"
            )
        )


class InvalidNormalization(CommandLineError):
    def __init__(self, norm):
        super().__init__(self)
        self.norm = norm

    def __str__(self):
        return self.render(
            (
                "\n"
                'You provided an invalid argument "%(norm)s" to normalize with. \n'
                'Please use "none" or "NFC", "NFKC", "NFD", or "NFKD"\n'
            )
        )


class MalformedMapping(CommandLineError):
    def __init__(self, message=""):
        super().__init__(self)
        if message:
            self.message = "\n\n" + message
        else:
            self.message = ""

    def __str__(self):
        return self.render(
            (
                "\n"
                "There is something wrong with your mapping. \n"
                "You might be missing some keywords or \n"
                'Not all of the input and output pairs in your mapping have values for "in" and "out"\n'
                'You may need to rerun "g2p update" after resolving this problem.\n'
                "Please refer to the documentation and fix your mapping." + self.message
            )
        )


class IncorrectFileType(CommandLineError):
    def __init__(self, msg):
        super().__init__(self)
        self.msg = msg

    def __str__(self):
        return self.render(self.msg)


class InvalidLanguageCode(CommandLineError):
    def __init__(self, lang):
        super().__init__(self)
        self.lang = lang

    def __str__(self):
        return self.render('No language called: "%(lang)s".')


class RecursionError(CommandLineError):
    def __init__(self, msg):
        super().__init__(self)
        self.msg = msg

    def __str__(self):
        return self.render(self.msg)
